##########################################################################
# Populse_mia - Copyright (C) IRMaGe/CEA, 2018
# Distributed under the terms of the CeCILL license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL_V2.1-en.html
# for details.
##########################################################################

import datetime
from traits.trait_base import Undefined
from traits.trait_handlers import TraitListObject

# Capsul imports
from capsul.api import Process

# Populse_MIA imports
from populse_mia.data_manager.project import COLLECTION_BRICK, BRICK_EXEC, \
    BRICK_EXEC_TIME, TAG_BRICKS, COLLECTION_INITIAL, COLLECTION_CURRENT, \
    BRICK_OUTPUTS


class ProcessMIA(Process):
    """
    Class overriding the default capsul Process class, in order to personalize
       the run for MIA bricks.

   This class is mainly used by MIA bricks.

    Methods:
        - _after_run_process: method called after the process being run
        - _before_run_process: method called before running the process
        - get_brick_to_update: give the brick to update given the scan list
           of bricks
        - get_scan_bricks: give the list of bricks given an output value
        - list_outputs: generates the outputs of the process (need to be
           overridden)
        - manage_brick_after_run: update process history after the run
           (Done status)
        - manage_brick_before_run: update process history before running
           the process
        - manage_brick_output_after_run: manage the bricks history before
           the run
        - manage_brick_output_before_run: manage the bricks history before
           the run
        - manage_matlab_launch_parameters: Set the Matlab's config parameters
           when a Nipype process is used
        - remove_brick_output: remove the bricks from the outputs

    """

    def __init__(self):
        super(ProcessMIA, self).__init__()
        # self.filters = {}  # use if the filters are set on plugs

    def _after_run_process(self, run_process_result):
        """Method called after the process being run.

        :param run_process_result: Result of the run process
        :return: the result of the run process
        """

        self.manage_brick_after_run()
        return run_process_result

    def _before_run_process(self):
        """Method called before running the process.

        Add the exec status Not Done and exec time to the process history
        """

        self.manage_brick_before_run()

    def get_brick_to_update(self, bricks):
        """Give the brick to update, given the scan list of bricks.

        :param bricks: list of scan bricks
        :return: Brick to update
        """

        if bricks is None:
            return

        if len(bricks) == 0:
            return None
        if len(bricks) == 1:
            return bricks[0]
        else:
            brick_to_keep = bricks[len(bricks) - 1]
            for brick in bricks:
                exec_status = self.project.session.get_value(COLLECTION_BRICK,
                                                             brick, BRICK_EXEC)
                exec_time = self.project.session.get_value(COLLECTION_BRICK,
                                                           brick,
                                                           BRICK_EXEC_TIME)
                if (exec_time is None and exec_status is None and brick !=
                        brick_to_keep):
                    # The other bricks not run are removed
                    outputs = self.project.session.get_value(COLLECTION_BRICK,
                                                             brick,
                                                             BRICK_OUTPUTS)
                    for output_name in outputs:
                        output_value = outputs[output_name]
                        self.remove_brick_output(brick, output_value)
                    self.project.session.remove_document(COLLECTION_BRICK,
                                                         brick)
                    self.project.saveModifications()
            return brick_to_keep

    def get_scan_bricks(self, output_value):
        """Give the list of bricks, given an output value.

        :param output_value: output value
        :return: list of bricks related to the output
        """
        for scan in self.project.session.get_documents_names(
                COLLECTION_CURRENT):
            if scan in str(output_value):
                return self.project.session.get_value(COLLECTION_CURRENT,
                                                      scan, TAG_BRICKS)
        return []

    def list_outputs(self):
        """Override the outputs of the process"""
        pass

    def manage_brick_after_run(self):
        """
        Manages the brick history after the run (Done status)
        """
        outputs = self.get_outputs()
        for output_name in outputs:
            output_value = outputs[output_name]
            if output_value not in ["<undefined>", Undefined]:
                if type(output_value) in [list, TraitListObject]:
                    for single_value in output_value:
                        self.manage_brick_output_after_run(single_value)
                else:
                    self.manage_brick_output_after_run(output_value)

    def manage_brick_before_run(self):
        """
        Updates process history, before running the process
        """

        outputs = self.get_outputs()
        for output_name in outputs:
            output_value = outputs[output_name]
            if output_value not in ["<undefined>", Undefined]:
                if type(output_value) in [list, TraitListObject]:
                    for single_value in output_value:
                        self.manage_brick_output_before_run(single_value)
                else:
                    self.manage_brick_output_before_run(output_value)

    def manage_brick_output_after_run(self, output_value):
        """
        Manages the bricks history before the run

        :param output_value: output value
        """

        scan_bricks_history = self.get_scan_bricks(output_value)
        brick_to_update = self.get_brick_to_update(scan_bricks_history)
        if brick_to_update is not None:
            self.project.session.set_value(COLLECTION_BRICK, brick_to_update,
                                           BRICK_EXEC, "Done")
            self.project.saveModifications()

    def manage_brick_output_before_run(self, output_value):
        """
        Manages the bricks history before the run

        :param output_value: output value
        """

        scan_bricks_history = self.get_scan_bricks(output_value)
        brick_to_update = self.get_brick_to_update(scan_bricks_history)
        if brick_to_update is not None:
            self.project.session.set_value(COLLECTION_BRICK, brick_to_update,
                                           BRICK_EXEC_TIME,
                                           datetime.datetime.now())
            self.project.session.set_value(COLLECTION_BRICK, brick_to_update,
                                           BRICK_EXEC, "Not Done")
            self.project.saveModifications()

    def manage_matlab_launch_parameters(self):
        """Set the Matlab's config parameters when a Nipype process is used

        Called in bricks.
        """

        if hasattr(self, "process"):
            self.process.inputs.use_mcr = self.use_mcr
            self.process.inputs.paths = self.paths
            self.process.inputs.matlab_cmd = self.matlab_cmd
            self.process.inputs.mfile = self.mfile

    def remove_brick_output(self, brick, output):
        """
        Removes the bricks from the outputs

        :param output: output
        :param brick: brick
        """

        if type(output) in [list, TraitListObject]:
            for single_value in output:
                self.remove_brick_output(brick, single_value)
            return

        for scan in self.project.session.get_documents_names(COLLECTION_CURRENT):
            if scan in output:
                output_bricks = self.project.session.get_value(
                    COLLECTION_CURRENT, scan, TAG_BRICKS)
                output_bricks.remove(brick)
                self.project.session.set_value(
                    COLLECTION_CURRENT, scan, TAG_BRICKS, output_bricks)
                self.project.session.set_value(
                    COLLECTION_INITIAL, scan, TAG_BRICKS, output_bricks)
                self.project.saveModifications()





# import os
# import sys
# import yaml
# import shutil
#
# # PyQt5 imports
# from PyQt5 import QtWidgets, QtGui, QtCore
#
# # Populse_MIA imports
# from populse_mia.software_properties import verCmp
#
#
# class MIAInstallWidget(QtWidgets.QWidget):
#
#     def __init__(self):
#
#         super().__init__()
#
#         # Labels
#         self.top_label_text = 'Welcome to Populse_MIA installation.'
#         self.top_label = QtWidgets.QLabel(self.top_label_text)
#
#         self.top_label_font = QtGui.QFont()
#         self.top_label_font.setBold(True)
#         self.top_label.setFont(self.top_label_font)
#
#         h_box_top_label = QtWidgets.QHBoxLayout()
#         h_box_top_label.addStretch(1)
#         h_box_top_label.addWidget(self.top_label)
#         h_box_top_label.addStretch(1)
#
#         self.middle_label_text = 'Please select an installation path and a folder to store your future projects.'
#
#         self.middle_label = QtWidgets.QLabel(self.middle_label_text)
#         h_box_middle_label = QtWidgets.QHBoxLayout()
#         h_box_middle_label.addStretch(1)
#         h_box_middle_label.addWidget(self.middle_label)
#         h_box_middle_label.addStretch(1)
#
#         # Groupbox
#         self.groupbox = QtWidgets.QGroupBox()
#
#         mia_path_default = ''  # to determine from the system (ex: for Windows: Program Files)
#
#         self.mia_path_label = QtWidgets.QLabel("Populse_MIA installation path:")
#         self.mia_path_choice = QtWidgets.QLineEdit(mia_path_default)
#         self.mia_path_browse = QtWidgets.QPushButton("Browse")
#         self.mia_path_browse.clicked.connect(self.browse_mia_path)
#
#         self.mia_path_info = QtWidgets.QPushButton(" ? ")
#         self.mia_path_info.setFixedHeight(27)
#         self.mia_path_info.setFixedWidth(27)
#         self.mia_path_info.setStyleSheet("background-color:rgb(150,150,200)")
#         rect = QtCore.QRect(4, 4, 17, 17)
#         region = QtGui.QRegion(rect, QtGui.QRegion.Ellipse)
#         self.mia_path_info.setMask(region)
#         tool_tip_message = "Two folders will be created in the selected folder:\n- populse_mia: containing " \
#                            "populse_mia's configuration and resources files.\n- MRIFileManager: containing " \
#                            "the file converter used in Populse_MIA."
#         self.mia_path_info.setToolTip(tool_tip_message)
#
#         h_box_mia_path = QtWidgets.QHBoxLayout()
#         h_box_mia_path.addWidget(self.mia_path_choice)
#         h_box_mia_path.addWidget(self.mia_path_browse)
#         h_box_mia_path.addWidget(self.mia_path_info)
#
#         v_box_mia_path = QtWidgets.QVBoxLayout()
#         v_box_mia_path.addWidget(self.mia_path_label)
#         v_box_mia_path.addLayout(h_box_mia_path)
#
#         projects_path_default = ''  # setting a default value for the projects?
#
#         self.projects_path_label = QtWidgets.QLabel("Populse_MIA projects path:")
#         self.projects_path_choice = QtWidgets.QLineEdit(projects_path_default)
#         self.projects_path_browse = QtWidgets.QPushButton("Browse")
#         self.projects_path_browse.clicked.connect(self.browse_projects_path)
#
#         self.projects_path_info = QtWidgets.QPushButton(" ? ")
#         self.projects_path_info.setFixedHeight(27)
#         self.projects_path_info.setFixedWidth(27)
#         self.projects_path_info.setStyleSheet("background-color:rgb(150,150,200)")
#         rect = QtCore.QRect(4, 4, 17, 17)
#         region = QtGui.QRegion(rect, QtGui.QRegion.Ellipse)
#         self.projects_path_info.setMask(region)
#         tool_tip_message = 'A "projects" folder will be created in this specified folder.'
#         self.projects_path_info.setToolTip(tool_tip_message)
#
#         h_box_projects_path = QtWidgets.QHBoxLayout()
#         h_box_projects_path.addWidget(self.projects_path_choice)
#         h_box_projects_path.addWidget(self.projects_path_browse)
#         h_box_projects_path.addWidget(self.projects_path_info)
#
#         v_box_projects_path = QtWidgets.QVBoxLayout()
#         v_box_projects_path.addWidget(self.projects_path_label)
#         v_box_projects_path.addLayout(h_box_projects_path)
#
#         v_box_paths = QtWidgets.QVBoxLayout()
#         v_box_paths.addLayout(v_box_mia_path)
#         v_box_paths.addLayout(v_box_projects_path)
#
#         self.groupbox.setLayout(v_box_paths)
#
#         # Push buttons
#         self.push_button_install = QtWidgets.QPushButton("Install")
#         self.push_button_install.clicked.connect(self.install)
#
#         self.push_button_cancel = QtWidgets.QPushButton("Cancel")
#         self.push_button_cancel.clicked.connect(self.close)
#
#         h_box_buttons = QtWidgets.QHBoxLayout()
#         h_box_buttons.addStretch(1)
#         h_box_buttons.addWidget(self.push_button_install)
#         h_box_buttons.addWidget(self.push_button_cancel)
#
#         # Final layout
#         self.global_layout = QtWidgets.QVBoxLayout()
#         self.global_layout.addLayout(h_box_top_label)
#         self.global_layout.addLayout(h_box_middle_label)
#         self.global_layout.addWidget(self.groupbox)
#         self.global_layout.addLayout(h_box_buttons)
#
#         self.setLayout(self.global_layout)
#         self.setWindowTitle("Populse_MIA installation")
#
#     def browse_mia_path(self):
#         folder_name = QtWidgets.QFileDialog.getExistingDirectory(self, 'Select a folder where to install Populse_MIA')
#         if folder_name:
#             self.mia_path_choice.setText(folder_name)
#
#     def browse_projects_path(self):
#         folder_name = QtWidgets.QFileDialog.\
#             getExistingDirectory(self, "Select a folder where to store Populse_MIA's projects")
#         if folder_name:
#             self.projects_path_choice.setText(folder_name)
#
#     def install(self):
#
#         self.project_flag = False
#
#         # Checking that the specified paths are correct
#         mia_path = self.mia_path_choice.text()
#         if not os.path.isdir(mia_path):
#             message = "The selected path for Populse_MIA must be an existing folder"
#             msg = QtWidgets.QMessageBox()
#             msg.setIcon(QtWidgets.QMessageBox.Warning)
#             msg.setText("Populse_MIA path is not valid")
#             msg.setInformativeText(message)
#             msg.setWindowTitle("Warning")
#             msg.setStandardButtons(QtWidgets.QMessageBox.Ok)
#             msg.buttonClicked.connect(msg.close)
#             msg.exec()
#             return
#
#         projects_path = self.projects_path_choice.text()
#         if not os.path.isdir(projects_path):
#             message = "The selected path for Populse_MIA's projects must be an existing folder"
#             msg = QtWidgets.QMessageBox()
#             msg.setIcon(QtWidgets.QMessageBox.Warning)
#             msg.setText("Populse_MIA's projects path is not valid")
#             msg.setInformativeText(message)
#             msg.setWindowTitle("Warning")
#             msg.setStandardButtons(QtWidgets.QMessageBox.Ok)
#             msg.buttonClicked.connect(msg.close)
#             msg.exec()
#             return
#
#         if os.path.isdir(os.path.join(projects_path, 'projects')):
#             message = 'A "projects" folder already exists in the selected path for the projects'
#             self.msg = QtWidgets.QMessageBox()
#             self.msg.setIcon(QtWidgets.QMessageBox.Warning)
#             self.msg.setText(message)
#             self.msg.setWindowTitle("Warning")
#             self.msg.setStandardButtons(QtWidgets.QMessageBox.Ok | QtWidgets.QMessageBox.Cancel)
#             # msg.buttonClicked.connect(msg.close)
#             self.msg.buttonClicked.connect(self.ok_or_abort)
#             self.msg.exec()
#
#         # If the user has clicked on "Cancel" the installation is aborted
#         if self.project_flag:
#             return
#
#         # Creating a "projects" folder in the specified projects folder
#         if not os.path.isdir(os.path.join(projects_path, 'projects')):
#             try:
#                 os.mkdir(os.path.join(projects_path, 'projects'))
#             except OSError as e:
#                 print('Error creating the "projects" folder: ', e)
#
#         # Moving populse_mia folder to the specified location
#         self.copy_directory('populse_mia', os.path.join(mia_path))
#
#         # Moving MRIFileManager folder to the specified location
#         self.copy_directory('MRIFileManager', os.path.join(mia_path))
#
#         # Adding both mia, MRIFileManager and projects paths to Populse_MIA's config
#         config_file = os.path.join(mia_path, 'properties', 'config.yml')
#         if os.path.isfile(config_file):
#             config_dic = self.load_config(config_file)
#             config_dic["mia_path"] = os.path.join(mia_path, 'populse_mia')
#             config_dic["projects_save_path"] = os.path.join(projects_path, 'projects')
#             config_dic["mri_conv_path"] = os.path.join(mia_path, 'MRIFileManager', 'MRIFileManager.jar')
#             self.save_config(config_dic, config_file)
#
#         # Installing Populse_MIA's modules using pip
#         self.install_package('populse_mia')  # Not available yet
#
#     def ok_or_abort(self, button):
#         role = self.msg.buttonRole(button)
#         if role == QtWidgets.QMessageBox.AcceptRole:  # "OK" has been clicked
#             self.project_flag = False
#         else:
#             self.project_flag = True
#
#     @staticmethod
#     def copy_directory(src, dest):
#         try:
#             shutil.copytree(src, dest)
#         # Directories are the same
#         except shutil.Error as e:
#             print('Directory not copied. Error: %s' % e)
#         # Any error saying that the directory doesn't exist
#         except OSError as e:
#             print('Directory not copied. Error: %s' % e)
#
#     @staticmethod
#     def load_config(config_file):
#         with open(config_file, 'r') as stream:
#
#             try:
#                 if verCmp(yaml.__version__, '5.1', 'sup'):
#                     return yaml.load(stream, Loader=yaml.FullLoader)
#
#                 else:
#                     return yaml.load(stream)
#
#             except yaml.YAMLError as exc:
#                 print(exc)
#
#     @staticmethod
#     def save_config(config_dic, config_file):
#         with open(config_file, 'w', encoding='utf8') as configfile:
#             yaml.dump(config_dic, configfile, default_flow_style=False, allow_unicode=True)
#
#     @staticmethod
#     def install_package(package):
#         import importlib
#         try:
#             importlib.import_module(package)
#         except ImportError:
#             import pip
#             pip.main(['install', package])
#
#
# if __name__ == '__main__':
#     app = QtWidgets.QApplication(sys.argv)
#     mia_install_widget = MIAInstallWidget()
#
#     # Setting the window to the middle of the screen
#     frame_gm = mia_install_widget.frameGeometry()
#     screen = QtWidgets.QApplication.desktop().screenNumber(QtWidgets.QApplication.desktop().cursor().pos())
#     center_point = QtWidgets.QApplication.desktop().screenGeometry(screen).center()
#     frame_gm.moveCenter(center_point)
#     mia_install_widget.move(frame_gm.topLeft())
#
#     mia_install_widget.show()
#     app.exec()
#



##########################################################################
# Populse_mia - Copyright (C) IRMaGe/CEA, 2018
# Distributed under the terms of the CeCILL license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL_V2.1-en.html
# for details.
##########################################################################

from datetime import datetime

# PyQt5 imports
from PyQt5.QtWidgets import (
    QDialog, QTableWidget, QVBoxLayout, QHBoxLayout,
    QTableWidgetItem, QPushButton, QMessageBox)

# Populse_MIA imports
from populse_mia.utils.utils import check_value_type
from populse_mia.data_manager.project import COLLECTION_CURRENT

# Populse_db imports
from populse_db.database import (
    FIELD_TYPE_LIST_INTEGER, FIELD_TYPE_LIST_FLOAT, FIELD_TYPE_LIST_STRING,
    FIELD_TYPE_LIST_DATE, FIELD_TYPE_LIST_DATETIME, FIELD_TYPE_LIST_TIME)


class ModifyTable(QDialog):
    """
    Verify precisely the scans of the project when called by the user.

    Methods:
        - fill_table: fill the table
        - update_table_values: update the table in the database
    """

    def __init__(self, project, value, types, scans, tags):
        """
        Initialization of the ModifyTable class

        :param project: Instance of project
        :param value: List of values of the cell
        :param types: Value types
        :param scans: Scans of the rows
        :param tags: Tags of the columns
        """
        super().__init__()

        self.setModal(True)

        # Variables init
        self.types = types
        self.scans = scans
        self.tags = tags
        self.project = project
        self.value = value

        # The table that will be filled
        self.table = QTableWidget()

        # Filling the table
        self.fill_table()

        # Ok button
        ok_button = QPushButton("Ok")
        ok_button.clicked.connect(self.update_table_values)

        # Cancel button
        cancel_button = QPushButton("Cancel")
        cancel_button.clicked.connect(self.close)

        # Layouts
        self.v_box_final = QVBoxLayout()
        self.h_box_final = QHBoxLayout()

        self.h_box_final.addWidget(ok_button)
        self.h_box_final.addWidget(cancel_button)

        self.v_box_final.addWidget(self.table)
        self.v_box_final.addLayout(self.h_box_final)

        self.setLayout(self.v_box_final)

    def fill_table(self):
        """Fill the table."""
        # Sizes
        self.table.setColumnCount(len(self.value))
        self.table.setRowCount(1)

        # Values filled
        for i in range (0, self.table.columnCount()):
            column_elem = self.value[i]
            item = QTableWidgetItem()
            item.setText(str(column_elem))
            self.table.setItem(0, i, item)

        # Resize
        self.table.resizeColumnsToContents()
        total_width = 0
        total_height = 0
        i = 0
        while i < self.table.columnCount():
            total_width += self.table.columnWidth(i)
            total_height += self.table.rowHeight(i)
            i += 1
        if total_width + 20 < 900:
            self.table.setFixedWidth(total_width + 20)
            self.table.setFixedHeight(total_height + 25)
        else:
            self.table.setFixedWidth(900)
            self.table.setFixedHeight(total_height + 40)

    def update_table_values(self):
        """Update the table in the database when the 'OK' button is clicked."""

        valid = True

        # For each value, type checked
        for i in range (0, self.table.columnCount()):
            item = self.table.item(0, i)
            text = item.text()

            valid_type = True
            for tag_type in self.types:
                if not check_value_type(text, tag_type, True):
                    valid_type = False
                    type_problem = tag_type
                    break

            # Type checked
            if not valid_type:

                # Error dialog if invalid cell
                valid = False
                msg = QMessageBox()
                msg.setIcon(QMessageBox.Warning)
                msg.setText("Invalid value")
                msg.setInformativeText("The value " + text +
                                       " is invalid with the type " +
                                       type_problem)
                msg.setWindowTitle("Warning")
                msg.setStandardButtons(QMessageBox.Ok)
                msg.buttonClicked.connect(msg.close)
                msg.exec()
                break

        if valid:
            # Database updated only if valid type for every cell

            for cell in range (0, len(self.scans)):
                scan = self.scans[cell]
                tag = self.tags[cell]
                tag_object = self.project.session.get_field(
                    COLLECTION_CURRENT, tag)
                tag_type = tag_object.type

                database_value = []

                # For each value
                for i in range (0, self.table.columnCount()):
                    item = self.table.item(0, i)
                    text = item.text()

                    if tag_type == FIELD_TYPE_LIST_INTEGER:
                        database_value.append(int(text))
                    elif tag_type == FIELD_TYPE_LIST_FLOAT:
                        database_value.append(float(text))
                    elif tag_type == FIELD_TYPE_LIST_STRING:
                        database_value.append(str(text))
                    elif tag_type == FIELD_TYPE_LIST_DATE:
                        format = "%d/%m/%Y"
                        subvalue = datetime.strptime(text, format).date()
                        database_value.append(subvalue)
                    elif tag_type == FIELD_TYPE_LIST_DATETIME:
                        format = "%d/%m/%Y %H:%M"
                        subvalue = datetime.strptime(text, format)
                        database_value.append(subvalue)
                    elif tag_type == FIELD_TYPE_LIST_TIME:
                        format = "%H:%M"
                        subvalue = datetime.strptime(text, format).time()
                        database_value.append(subvalue)

                # Database updated for every cell
                self.project.session.set_value(
                    COLLECTION_CURRENT, scan, tag, database_value)

            self.close()



##########################################################################
# Populse_mia - Copyright (C) IRMaGe/CEA, 2018
# Distributed under the terms of the CeCILL license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL_V2.1-en.html
# for details.
##########################################################################

import ast
import os
import re
from datetime import datetime, date, time
import dateutil.parser

# PyQt5 imports
from PyQt5.QtCore import Qt, QVariant, QDateTime, QTime, QDate
from PyQt5.QtWidgets import QMessageBox

# Populse_db imports
from populse_db.database import (
    FIELD_TYPE_INTEGER, FIELD_TYPE_LIST_INTEGER, FIELD_TYPE_STRING,
    FIELD_TYPE_BOOLEAN, FIELD_TYPE_FLOAT, FIELD_TYPE_TIME, FIELD_TYPE_DATE,
    FIELD_TYPE_DATETIME, FIELD_TYPE_LIST_TIME, FIELD_TYPE_LIST_DATETIME,
    FIELD_TYPE_LIST_DATE, LIST_TYPES, FIELD_TYPE_LIST_FLOAT,
    FIELD_TYPE_LIST_BOOLEAN, FIELD_TYPE_LIST_STRING)

from populse_mia.software_properties import Config


def check_value_type(value, value_type, is_subvalue=False):
    """
    Checks the type of a new value

    :param value: Value of the cell
    :param value_type: Type expected
    :param is_subvalue: To know if the value is a subvalue of a list
    :return: True if the value is valid to replace the old one, False otherwise
    """
    if (value_type == FIELD_TYPE_INTEGER or
            value_type == FIELD_TYPE_LIST_INTEGER and is_subvalue):
        try:
            int(value)
            return True
        except Exception:
            return False
    elif (value_type == FIELD_TYPE_FLOAT or
          value_type == FIELD_TYPE_LIST_FLOAT and is_subvalue):
        try:
            float(value)
            return True
        except Exception:
            return False
    elif (value_type == FIELD_TYPE_BOOLEAN or
        value_type == FIELD_TYPE_LIST_BOOLEAN and is_subvalue):
        return (value == "True" or value == True or value == "False" or
                value == False)
    elif (value_type == FIELD_TYPE_STRING or
            value_type == FIELD_TYPE_LIST_STRING and is_subvalue):
        try:
            str(value)
            return True
        except Exception:
            return False
    elif value_type in LIST_TYPES and not is_subvalue:
        if isinstance(value, str):
            value = ast.literal_eval(value)
        is_valid_value = True
        for subvalue in value:
            if not check_value_type(subvalue, value_type, True):
                is_valid_value = False
                break
        return is_valid_value
    elif (value_type == FIELD_TYPE_DATE or
            value_type == FIELD_TYPE_LIST_DATE and is_subvalue):
        if isinstance(value, QDate):
            return True
        elif isinstance(value, str):
            try:
                format = "%d/%m/%Y"
                datetime.strptime(value, format).date()
                return True
            except Exception:
                return False
    elif (value_type == FIELD_TYPE_DATETIME or
            value_type == FIELD_TYPE_LIST_DATETIME and is_subvalue):
        if isinstance(value, QDateTime):
            return True
        elif isinstance(value, str):
            try:
                format = "%d/%m/%Y %H:%M:%S.%f"
                datetime.strptime(value, format)
                return True
            except Exception:
                return False
    elif (value_type == FIELD_TYPE_TIME or
          value_type == FIELD_TYPE_LIST_TIME and is_subvalue):
        if isinstance(value, QTime):
            return True
        elif isinstance(value, str):
            try:
                format = "%H:%M:%S.%f"
                datetime.strptime(value, format).time()
                return True
            except Exception:
                return False


def message_already_exists():
    """
    Displays a message box to tell that a name already exists
    """
    msg = QMessageBox()
    msg.setIcon(QMessageBox.Warning)
    msg.setText("This name already exists in this parent folder")
    msg.setWindowTitle("Warning")
    msg.setStandardButtons(QMessageBox.Ok)
    msg.buttonClicked.connect(msg.close)
    msg.exec()


def set_filters_directory_as_default(dialog):
    """
    Sets the filters directory as default (Json files)

    :param dialog: current file dialog
    """
    if not (os.path.exists(os.path.join(
            os.path.join(os.path.relpath(
                os.curdir), '..', '..'), 'filters'))):
        os.makedirs(os.path.join(
            os.path.join(os.path.relpath(
                os.curdir), '..', '..'), 'filters'))
    dialog.setDirectory(
        os.path.expanduser(os.path.join(
            os.path.join(os.path.relpath(
                os.curdir), '..', '..'), 'filters')))


def set_item_data(item, value, value_type):
    """
    Sets the item data in the data browser

    :param item: item to set
    :param value: new item value
    :param value_type: new value type
    """

    if value_type in LIST_TYPES:
        if isinstance(value, str):
            value = ast.literal_eval(value)
        if value_type == FIELD_TYPE_LIST_DATE:
            new_list = []
            for subvalue in value:
                new_list.append(subvalue.strftime('%d/%m/%Y'))
            value = new_list
        elif value_type == FIELD_TYPE_LIST_DATETIME:
            new_list = []
            for subvalue in value:
                new_list.append(subvalue.strftime('%d/%m/%Y %H:%M:%S.%f'))
            value = new_list
        elif value_type == FIELD_TYPE_LIST_TIME:
            new_list = []
            for subvalue in value:
                new_list.append(subvalue.strftime('%H:%M:%S.%f'))
            value = new_list
        value_prepared = str(value)
        item.setData(Qt.EditRole, QVariant(value_prepared))
    elif value_type == FIELD_TYPE_DATETIME:
        if isinstance(value, datetime):
            value_prepared = QDateTime(value)
        elif isinstance(value, QDateTime):
            value_prepared = value
        elif isinstance(value, str):
            format = "%d/%m/%Y %H:%M:%S.%f"
            value_prepared = QDateTime(datetime.strptime(value, format))
        item.setData(Qt.EditRole, QVariant(value_prepared))
    elif value_type == FIELD_TYPE_DATE:
        if isinstance(value, date):
            value_prepared = QDate(value)
        elif isinstance(value, QDate):
            value_prepared = value
        elif isinstance(value, str):
            format = "%d/%m/%Y"
            value_prepared = QDate(datetime.strptime(value, format).date())
        item.setData(Qt.EditRole, QVariant(value_prepared))
    elif value_type == FIELD_TYPE_TIME:
        if isinstance(value, time):
            value_prepared = QTime(value)
        elif isinstance(value, QTime):
            value_prepared = value
        elif isinstance(value, str):
            format = "%H:%M:%S.%f"
            value_prepared = QTime(datetime.strptime(value, format).time())
        item.setData(Qt.EditRole, QVariant(value_prepared))
    elif value_type == FIELD_TYPE_FLOAT:
        value_prepared = float(value)
        item.setData(Qt.EditRole, QVariant(value_prepared))
    elif value_type == FIELD_TYPE_INTEGER:
        value_prepared = int(value)
        item.setData(Qt.EditRole, QVariant(value_prepared))
    elif value_type == FIELD_TYPE_BOOLEAN:
        value_prepared = value
        item.setData(Qt.EditRole, QVariant(value_prepared))
    elif value_type == FIELD_TYPE_STRING:
        value_prepared = str(value)
        item.setData(Qt.EditRole, QVariant(value_prepared))


def set_projects_directory_as_default(dialog):
    """
    Sets the projects directory as default

    :param dialog: current file dialog
    """
    config = Config()
    projects_directory = config.get_projects_save_path()
    if not os.path.exists(projects_directory):
        os.makedirs(projects_directory)
    dialog.setDirectory(projects_directory)


def table_to_database(value, value_type):
    """
    Prepares the value to the database

    :param value: Value to convert
    :param value_type: Value type
    :return: The value converted for the database
    """

    if value_type == FIELD_TYPE_FLOAT:
        return float(value)
    elif value_type == FIELD_TYPE_STRING:
        return str(value)
    elif value_type == FIELD_TYPE_INTEGER:
        return int(value)
    elif value_type == FIELD_TYPE_BOOLEAN:
        if value == "True" or value == True:
            return True
        elif value == "False" or value == False:
            return False
    elif value_type == FIELD_TYPE_DATETIME:
        if isinstance(value, QDateTime):
            return value.toPyDateTime()
        elif isinstance(value, str):
            try:
                format = "%d/%m/%Y %H:%M:%S.%f"
                date_typed = datetime.strptime(value, format)
            except Exception:
                date_typed = dateutil.parser.parse(value)
            return date_typed

    elif value_type == FIELD_TYPE_DATE:
        if isinstance(value, QDate):
            return value.toPyDate()
        elif isinstance(value, str):
            format = "%d/%m/%Y"
            return datetime.strptime(value, format).date()
    elif value_type == FIELD_TYPE_TIME:
        if isinstance(value, QTime):
            return value.toPyTime()
        elif isinstance(value, str):
            format = "%H:%M:%S.%f"
            return datetime.strptime(value, format).time()
    elif value_type in LIST_TYPES:
        old_list = ast.literal_eval(value)
        list_to_return = []
        for old_element in old_list:
            list_to_return.append(table_to_database(
                old_element, value_type.replace("list_", "")))
        return list_to_return





# -*- coding: utf-8 -*- # Character encoding, recommended
"""The first module used at the runtime of mia.

Basically, this module is dedicated to the initialisation of the basic
parameters and the various checks necessary for a successful launch of the
mia's GUI.

:Contains:
    :Class:
        - PackagesInstall

    :Function:

        - launch_mia
        - main
        - verify_processes

"""

###############################################################################
# Populse_mia - Copyright (C) IRMaGe/CEA, 2018
# Distributed under the terms of the CeCILL license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL_V2.1-en.html
# for details.
###############################################################################

import sys
import os
import pkgutil
import inspect
import yaml
import copy
from functools import partial

# PyQt5 imports
from PyQt5.QtCore import QDir, QLockFile, Qt
from PyQt5.QtWidgets import (QApplication, QDialog, QPushButton, QLabel,
                             QFileDialog, QVBoxLayout, QHBoxLayout, QLineEdit)

# soma-base imports
from soma.qt_gui.qt_backend.Qt import QMessageBox

# Adding populse_mia path to the sys.path if in developer mode
if not os.path.dirname(os.path.dirname(
        os.path.realpath(__file__))) in sys.path:           # "developer" mode
    print('\nPopulse_MIA in "developer" mode')
    sys.path.insert(0, os.path.dirname(os.path.dirname(
        os.path.realpath(__file__))))
    DEV_MODE = True

else:                                                       # "user" mode
    DEV_MODE = False

# populse_mia imports
from populse_mia.user_interface.main_window import MainWindow
from populse_mia.data_manager.project import Project
from populse_mia.software_properties import Config
from populse_mia.software_properties import verCmp
from populse_mia.data_manager.project_properties import SavedProjects

main_window = None


class PackagesInstall:
    
    """Help to make available a pipeline package in the mia pipeline library,
in a recursive way.

    :Contains:
        :Method:
            - __init__
            - add_package

    """

    def __init__(self):
        """Initialise the packages instance attribute."""
        self.packages = {}

    def add_package(self, module_name, class_name=None):
        """Provide recursive representation of a package and its
subpackages/modules, to construct the mia's pipeline library.

        :Parameters:
            - :module_name: name of the module to add in the pipeline
               library
            - :class_name: only this pipeline will be add to the pipeline
               library (optional)

        :returns: dictionary of dictionaries containing
           package/subpackages/pipelines status.
           ex: {package: {subpackage: {pipeline: 'process_enabled'}}}

        """

        if module_name:

            # reloading the package
            if module_name in sys.modules.keys():
                del sys.modules[module_name]

            try:
                __import__(module_name)
                pkg = sys.modules[module_name]

                # check if there are subpackages, in this case explore them
                for _, modname, ispkg in pkgutil.iter_modules(pkg.__path__):

                    if ispkg:
                        print('\nExploring subpackages  of {0} ...'
                              .format(module_name))
                        print('- ', str(module_name + '.' + modname))
                        self.add_package(str(module_name + '.' + modname),
                                         class_name)

                for k, v in sorted(list(pkg.__dict__.items())):

                    if class_name and k != class_name:
                        continue

                    # checking each class in the package
                    if inspect.isclass(v):

                        # updating the tree's dictionary
                        path_list = module_name.split('.')
                        path_list.append(k)
                        pkg_iter = self.packages

                        for element in path_list:

                            if element in pkg_iter.keys():
                                pkg_iter = pkg_iter[element]

                            else:

                                if element is path_list[-1]:
                                    pkg_iter[element] = 'process_enabled'

                                else:
                                    pkg_iter[element] = {}
                                    pkg_iter = pkg_iter[element]

            except ImportError as e:
                print('\nWhen attempting to add a package and its modules to '
                      'the package tree, the following exception was caught:')
                print('{0}'.format(e))

            return self.packages


def launch_mia():
    """Actual launch of the mia software.

    Overload the sys.excepthook handler with the _my_excepthook private
    function. Check if the software is already opened in another instance.
    If not, the list of opened projects is cleared. Checks if saved projects
    known in the mia software still exist, and updates if necessary.
    Instantiates a 'project' object that handles projects and their
    associated database anf finally launch of the mia's GUI.


    **Contains: Private function:**
        - *_my_excepthook(etype, evalue, tback)*
           Log all uncaught exceptions in non-interactive mode.

           All python exceptions are handled by function, stored in
           sys.excepthook. By overloading the sys.excepthook handler with
           _my_excepthook function, this last function is called whenever
           there is a unhandled exception (so one that exits the interpreter).
           We take advantage of it to clean up mia software before closing.

            :Parameters:
                - :etype: exception class
                - :evalue: exception instance
                - :tback: traceback object

            **Contains: Private function:**
              - *_clean_up()*
                 Cleans up the mia software during "normal" closing.

                 Make a clean up of the opened projects just before exiting
                 mia.
        - *_verify_saved_projects()*
               Verify if the projects saved in saved_projects.yml are still
               on the disk.

               :Returns: the list of the deleted projects


    """

    def _my_excepthook(etype, evalue, tback):

        def _clean_up():

            global main_window
            config = Config()
            opened_projects = config.get_opened_projects()

            try:
                opened_projects.remove(main_window.project.folder)
                config.set_opened_projects(opened_projects)
                main_window.remove_raw_files_useless()

            except AttributeError as e:
                opened_projects = []
                config.set_opened_projects(opened_projects)

            print("\nClean up before closing mia done ...\n")

        # log the exception here
        print("\nException hooking in progress ...")
        _clean_up()
        # then call the default handler
        sys.__excepthook__(etype, evalue, tback)
        # there was some issue/error/problem, so exiting
        sys.exit(1)

    def _verify_saved_projects():

        saved_projects_object = SavedProjects()
        saved_projects_list = copy.deepcopy(saved_projects_object.pathsList)
        deleted_projects = []
        for saved_project in saved_projects_list:
            if not os.path.isdir(saved_project):
                deleted_projects.append(saved_project)
                saved_projects_object.removeSavedProject(saved_project)

        return deleted_projects

    global main_window
    app = QApplication(sys.argv)
    QApplication.setOverrideCursor(Qt.WaitCursor)
    sys.excepthook = _my_excepthook

    # working from the scripts directory
    os.chdir(os.path.dirname(os.path.realpath(__file__)))
    lock_file = QLockFile(QDir.temp().absoluteFilePath(
        'lock_file_populse_mia.lock'))

    if not lock_file.tryLock(100):
        # software already opened in another instance
        pass

    else:
        # no instances of the software is opened, the list of opened projects
        # can be cleared
        config = Config()
        config.set_opened_projects([])

    deleted_projects = _verify_saved_projects()
    project = Project(None, True)
    main_window = MainWindow(project, deleted_projects=deleted_projects)
    main_window.show()
    app.exec()


def check_python_version():
    if sys.version_info[:2] < (3, 5):
        raise AssertionError("The populse_mia is ensured to work with Python "
                             ">= 3.5")


def main():
    """Make basic configuration check then actual launch of mia.

    Checks if MIA is called from the site/dist packages (user mode) or from a
    cloned git repository (developer mode). Tries to update the dev_mode
    parameter accordingly and the mia_path if necessary, in the
    ~/.populse_mia/configuration.yml file (this file must be available from
    the ~/.populse_mia directory). Launches the verify_processes()
    function, then the launch_mia() function (mia's real launch !!). When
    mia is exited, if the ~/.populse_mia/configuration.yml exists, sets the
    dev_mode parameter to False.

    - If launched from a cloned git repository ('developer mode'):
        - if the ~/.populse_mia/configuration.yml exists, updates
          the dev_mode parameter to True
        - if the ~/.populse_mia/configuration.yml is not existing
          nothing is done (in developer mode, the mia_path is the
          cloned git repository.
    - If launched from the site/dist packages ('user mode'):
        - if the ~/.populse_mia/configuration.yml exists, updates
          the dev_mode parameter to False and, if not found, update
          the mia_path parameter.
        - if the file ~/.populse_mia/configuration.yml file does
          not exist or if the returned mia_path parameter is incorrect, a
          valid mia_path path is requested from the user, in order
          to try to fix a corruption of this file.

    :Contains:
        :Private function:
            - *_browse_mia_path()*
                The user define the mia_path parameter.

                This method, used only if the mia configuration parameters are
                not accessible, goes with the _verify_miaConfig function, which 
                will use the value of the mia_path parameter, defined here. 
                
                :Parameters dialog: QtWidgets.QDialog object ('msg' in the main
                   function)

            - *_verify_miaConfig()*

                The purpose of this method is twofold. First, it allows to
                update the obsolete values for some parameters of the
                mia_user_path/properties/config.yml file. Secondly, it allows
                to correct the value of the mia_user_path parameter in the
                ~/.populse_mia/configuration.yml file, (in the case of a user
                mode launch, because the developer mode launch does not need
                this parameter).

                In the case of a launch in user mode, this method goes with the
                _browse_mia_path() function, the latter having allowed the
                definition of the mia_user_path parameter, the objective here
                is to check if the value of this parameter is valid. The
                dev_mode=False and mia_path parameters are saved in the
                ~/.populse_mia/configuration.yml file (the mia_user_path
                parameter is mandatory in the user mode). Then the data in the
                mia_path/properties/config.yml file are checked. If an exception
                is raised during the  _verify_miaConfig function, the "MIA path
                selection" window is not closed and the user is prompted again
                to set the mia_user_path parameter.

                :Parameters dialog: QtWidgets.QDialog object ('msg' in the main
                   function)

    """

    def _browse_mia_path(dialog):

        dname = QFileDialog.getExistingDirectory(dialog,
                                                 "Please select MIA path",
                                                 os.path.expanduser('~'))
        dialog.file_line_edit.setText(dname)

    def _verify_miaConfig(dialog=None):

        save_flag = False
        
        if DEV_MODE:                     # "developer" mode

            try:
                config = Config()

            except Exception as e:
                print('\nMIA configuration settings could not be '
                      'recovered: {0} ...'.format(e))
                print('\nMIA is exiting ...\n')
                sys.exit(1)
                    
        elif dialog is not None:         # "user" mode only if problem
            
            mia_home_config = dict()
            mia_home_config["dev_mode"] = False
            mia_home_config["mia_user_path"] = dialog.file_line_edit.text()
            print('\nNew values in ~/.populse_mia/configuration.yml: ')

            for key, value in mia_home_config.items():
                print('- {0}: {1}'.format(key, value))

            print()
                      
            with open(dot_mia_config, 'w', encoding='utf8') as configfile:
                yaml.dump(mia_home_config, configfile, default_flow_style=False,
                          allow_unicode=True)

            try:
                config = Config()
                dialog.close()

            except Exception as e:
                print('\nCould not fetch the '
                      'configuration file: {0} ...'.format(e, ))
                msg = QMessageBox()
                msg.setIcon(QMessageBox.Warning)
                msg.setWindowTitle("populse_mia - Error: "
                                   "mia path directory incorrect")
                msg.setText("Error: Please select the MIA path (directory with"
                            "\nthe processes, properties & resources "
                            "directories): ")
                msg.setStandardButtons(QMessageBox.Ok)
                msg.buttonClicked.connect(msg.close)
                msg.exec()

        else:                            # "user" mode (initial pass)
            config = Config()

        if 'config' in locals():
            
            for key, value in config.config.items():
                
                if value == 'no':
                    save_flag = True
                    config.config[key] = False
                    
                if value == 'yes':
                    save_flag = True
                    config.config[key] = True

                if save_flag is True:
                    config.saveConfig()

    dot_mia_config = os.path.join(os.path.expanduser("~"), ".populse_mia",
                                  "configuration.yml")

    if DEV_MODE:                         # "developer" mode

        if os.path.isfile(dot_mia_config):
            print('\n{0} has been detected.'.format(dot_mia_config))

            with open(dot_mia_config, 'r') as stream:
                
                if verCmp(yaml.__version__, '5.1', 'sup'):
                    mia_home_config = yaml.load(stream,
                                                Loader=yaml.FullLoader)
                else:    
                    mia_home_config = yaml.load(stream)

            mia_home_config["dev_mode"] = True

            with open(dot_mia_config, 'w', encoding='utf8') as configfile:
                yaml.dump(mia_home_config,
                          configfile,
                          default_flow_style=False,
                          allow_unicode=True)

        else:
            print('\n{0} has not been detected.'.format(dot_mia_config))

        _verify_miaConfig()

    else:                                # "user" mode
  
        try:
            
            if not os.path.exists(os.path.dirname(dot_mia_config)):
                os.mkdir(os.path.dirname(dot_mia_config))
                print('\nThe {0} directory is created ...'
                      .format(os.path.exists(os.path.dirname(dot_mia_config))))

            with open(dot_mia_config, 'r') as stream:
                
                if verCmp(yaml.__version__, '5.1', 'sup'):
                    mia_home_config = yaml.load(stream,
                                                Loader=yaml.FullLoader)
                else:    
                    mia_home_config = yaml.load(stream)


            mia_home_config["dev_mode"] = False
            
            with open(dot_mia_config, 'w', encoding='utf8') as configfile:
                yaml.dump(mia_home_config, configfile,
                          default_flow_style=False,
                          allow_unicode=True)
                
            _verify_miaConfig()

        except Exception as e:  # the configuration.yml file does not exist
            # or has not been correctly read ...
            print('\nA probleme has been detected when opening'
                  ' the ~/.populse_mia/configuration.yml file'
                  ' or with the parameters returned from this file: ', e)
            mia_home_config = dict()
            mia_home_config["dev_mode"] = False

            # open popup, user choose the path to .populse_mia/populse_mia
            app = QApplication(sys.argv)
            msg = QDialog()
            msg.setWindowTitle("populse_mia - mia path selection")
            vbox_layout = QVBoxLayout()
            hbox_layout = QHBoxLayout()
            file_label = QLabel("Please select the MIA path (directory with\n "
                                "the processes, properties & resources "
                                "directories): ")
            msg.file_line_edit = QLineEdit()
            msg.file_line_edit.setFixedWidth(400)
            file_button = QPushButton("Browse")
            file_button.clicked.connect(partial(_browse_mia_path, msg))
            vbox_layout.addWidget(file_label)
            hbox_layout.addWidget(msg.file_line_edit)
            hbox_layout.addWidget(file_button)
            vbox_layout.addLayout(hbox_layout)
            hbox_layout = QHBoxLayout()
            msg.ok_button = QPushButton("Ok")
            msg.ok_button.clicked.connect(partial(_verify_miaConfig, msg))
            hbox_layout.addWidget(msg.ok_button)
            vbox_layout.addLayout(hbox_layout)
            msg.setLayout(vbox_layout)
            msg.exec()
            del app
      
    verify_processes()
    check_python_version()
    launch_mia()

    # set the dev_mode to False when exiting mia,
    # if ~/.populse_mia/configuration.yml file exists
    if os.path.isfile(dot_mia_config):
        
        with open(dot_mia_config, 'r') as stream:
            
            if verCmp(yaml.__version__, '5.1', 'sup'):
                mia_home_config = yaml.load(stream,
                                            Loader=yaml.FullLoader)
            else:    
                mia_home_config = yaml.load(stream)

        mia_home_config["dev_mode"] = False

        with open(dot_mia_config, 'w', encoding='utf8') as configfile:
            yaml.dump(mia_home_config, configfile, default_flow_style=False,
                      allow_unicode=True)


def verify_processes():
    """Install or update to the last version available on the station, of the \
       nipype and the mia_processes processes libraries.

    By default, mia provides two process libraries in the pipeline library
    (available in Pipeline Manager tab). The nipype, given as it is because
    it is developed by another team (https://github.com/nipy/nipype), and
    mia_processes which is developed under the umbrella of populse
    (https://github.com/populse/mia_processes). When installing mia in
    user mode, these two libraries are automatically installed on the
    station. The idea is to use the versioning available with pypi
    (https://pypi.org/). Thus, it is sufficient for the user to change the
    version of the library installed on the station (pip install...) to
    also change the version available in mia. Indeed, when starting mia, the
    verify_processes function will install or update nipype and
    mia_processes libraries in the pipeline library. Currently it is
    mandatory to have nipype and may_processes installed in the station.
    All these informations, as well as the installed versions and package
    paths are saved in the  mia_path/properties/process_config.yml file.
    When an upgrade or downgrade is performed for a package, the last
    configuration used by the user is kept (if a pipeline was visible, it
    remains so and vice versa). However, if a new pipeline is available in
    the new version it is automatically marked as visible in the library.

    :Contains:
        :Private function:
            - *_deepCompDic()*

               Try to keep the previous configuration existing before the
               update of the packages.

               Recursive comparison of the old_dic and new _dic dictionary. If
               all keys are recursively identical, the final value at the end
               of the whole tree in old_dic is kept in the new _dic. To sum
               up, this function is used to keep up the user display
               preferences in the processes library of the Pipeline Manager
               Editor.

               **param old_dic**: the dic representation of the previous
                 package configuration

               **param new_dic**: the dic representation of the new package \
                  configuration

               **param level**: the index level in the package (0: root,
                   +1: in a subpackage/pipeline)

               :returns: True if the current level is a pipeline that existed
                   in the old configuration, False if the
                   package/subpackage/pipeline did not exist

    """

    def _deepCompDic(old_dic, new_dic, level="0"):

        PY3 = sys.version_info[0] == 3

        if PY3:

            if isinstance(old_dic, str):
                return True

        else:
            if isinstance(old_dic, basestring):
                return True

        for key in old_dic:

            if key not in new_dic:

                if level == "0":
                    pass

                elif level == "+1":
                    return False

            # keep the same configuration for the pipeline in new and old dic
            elif (_deepCompDic(old_dic[str(key)], new_dic[str(key)],
                               level="+1")):
                new_dic[str(key)] = old_dic[str(key)]

    proc_content = False
    nipypeVer = False
    miaProcVer = False
    pack2install = []
    config = Config()
    proc_config = os.path.join(config.get_mia_path(), 'properties',
                               'process_config.yml')
    # check if nipype and mia_processes are available on the station
    # if not available inform the user to install them
    print('\nChecking the installed versions of nipype and mia_processes ...')
    print('***************************************************************')
    pkg_error = []
    
    try:
        __import__('nipype')
        nipypeVer = sys.modules['nipype'].__version__
        
    except ImportError as e:
        pkg_error.append('nipype')
        print('\n' + '*' * 37)
        print('MIA warning: {0}'.format(e))
        print('*' * 37 + '\n')

    try:
        __import__('mia_processes')
        miaProcVer = sys.modules['mia_processes'].__version__

    except ImportError as e:
        pkg_error.append('mia_processes')
        print('\n' + '*' * 37)
        print('MIA warning: {0}'.format(e))
        print('*' * 37 + '\n')
   
    if len(pkg_error) > 0:
        app = QApplication(sys.argv)
        msg = QMessageBox()
        msg.setIcon(QMessageBox.Warning)
        msg.setWindowTitle("populse_mia -  warning: ModuleNotFoundError!")

        if len(pkg_error) == 1:
            msg.setText("{0} package not found !\nPlease install "
                        "the package and "
                        "start again mia ...".format(pkg_error[0]))
        else:
            msg.setText("{0} and {1} packages not found !\n"
                        "Please install the packages and start again mia "
                        "...".format(pkg_error[0], pkg_error[1]))
        msg.setStandardButtons(QMessageBox.Ok)
        msg.buttonClicked.connect(msg.close)
        msg.exec()
        del app
        sys.exit(1)

    if os.path.isfile(proc_config):
        
        with open(proc_config, 'r') as stream:

            if verCmp(yaml.__version__, '5.1', 'sup'):
                proc_content = yaml.load(stream,
                                         Loader=yaml.FullLoader)
            else:    
                proc_content = yaml.load(stream)

    if (isinstance(proc_content, dict)) and ('Packages' in proc_content):
        othPckg = [f for f in proc_content['Packages']
                   if f not in ['mia_processes', 'nipype']]

    if 'othPckg' in dir():
        # othPckg: a list containing all packages, other than nipype and
        # mia_processes, used during the previous launch of mia.
        
        for pckg in othPckg:

            try:
                __import__(pckg)

            except ImportError as e:
                # try to update the sys.path for the processes/ directory
                # currently used
                if (not os.path.relpath(os.path.join(config.get_mia_path(),
                                                     'processes')) in
                        sys.path) and (not os.path.abspath(os.path.join(
                        config.get_mia_path(), 'processes')) in sys.path):
                    sys.path.append(os.path.abspath(os.path.join(
                        config.get_mia_path(), 'processes')))

                    try:
                        __import__(pckg)

                        # update the Paths parameter (processes/ directory
                        # currently used) saved later in the
                        # mia_path/properties/process_config.yml file
                        if (('Paths' in proc_content)
                                and (isinstance(proc_content['Paths'], list))):

                            if ((not os.path.relpath(os.path.join(
                                    config.get_mia_path(), 'processes')) in
                                     proc_content['Paths'])
                                    and (
                                            not os.path.abspath(os.path.join(
                                                config.get_mia_path(),
                                                'processes')) in
                                                proc_content['Paths'])):
                                proc_content['Paths'].append(os.path.abspath(
                                    os.path.join(
                                        config.get_mia_path(), 'processes')))

                        else:
                            proc_content['Paths'] = [os.path.abspath(
                                os.path.join(config.get_mia_path(),
                                             'processes'))]

                        with open(proc_config, 'w', encoding='utf8') as stream:
                            yaml.dump(proc_content, stream,
                                      default_flow_style=False,
                                      allow_unicode=True)

                        # finally the processes/ directory currently used is
                        # removed from the sys.path because this directory is
                        # now added to the Paths parameter in the
                        # mia_path/properties/process_config.yml file
                        sys.path.remove(os.path.abspath(os.path.join(
                            config.get_mia_path(), 'processes')))

                    # if an exception is raised, ask to the user to remove the
                    # package from the pipeline library or reload it
                    except ImportError as e:
                        print('{0}'.format(e))
                        app = QApplication(sys.argv)
                        msg = QMessageBox()
                        msg.setIcon(QMessageBox.Warning)
                        msg.setWindowTitle("populse_mia - warning: {0}"
                                           .format(e))
                        msg.setText(("At least, {0} has not been found in {1}."
                                     "\nTo prevent mia crash when using it, "
                                     "please remove (see File > Package "
                                     "library manager) or load again (see More"
                                     " > Install processes) the corresponding "
                                     "process library.").format(
                            e.msg.split()[-1], os.path.abspath(os.path.join(
                                config.get_mia_path(), 'processes'))))
                        msg.setStandardButtons(QMessageBox.Ok)
                        msg.buttonClicked.connect(msg.close)
                        msg.exec()
                        del app
                        sys.path.remove(os.path.abspath(os.path.join(
                            config.get_mia_path(), 'processes')))

                # the processes/ directory being already in the sys.path, the
                # package is certainly not properly installed in the processes/
                # directory
                else:
                    print("No module named '{0}'".format(pckg))
                    app = QApplication(sys.argv)
                    msg = QMessageBox()
                    msg.setIcon(QMessageBox.Warning)
                    msg.setWindowTitle("populse_mia - warning: {0}".format(e))
                    msg.setText(("At least, {0} has not been found in {1}."
                                 "\nTo prevent mia crash when using it, "
                                 "please remove (see File > Package "
                                 "library manager) or load again (see More"
                                 " > Install processes) the corresponding "
                                 "process library.").format(
                        e.msg.split()[-1], os.path.abspath(os.path.join(
                            config.get_mia_path(), 'processes'))))
                    msg.setStandardButtons(QMessageBox.Ok)
                    msg.buttonClicked.connect(msg.close)
                    msg.exec()
                    del app

    # the file mia_path/properties/process_config.yml is corrupted or
    # no pipeline processes was available during the previous use of mia or
    # their version is not known
    if (not isinstance(proc_content, dict)) or (
            (isinstance(proc_content, dict)) and
            ('Packages' not in proc_content)) or (
            (isinstance(proc_content, dict)) and
            ('Versions' not in proc_content)):
        pack2install = ['nipype.interfaces', 'mia_processes']

    else:
        # during the previous use of mia, nipype was not available or its
        # version was not known or its version was different from the one
        # currently available on the station
        if ((isinstance(proc_content, dict)) and ('Packages' in proc_content)
            and ('nipype' not in proc_content['Packages'])) or (
                (isinstance(proc_content, dict)) and ('Versions' in
                                                      proc_content)
                and ('nipype' not in proc_content['Versions'])) or (
                (isinstance(proc_content, dict)) and ('Versions' in
                                                      proc_content)
                and ('nipype' in proc_content['Versions'])
                and (proc_content['Versions']['nipype'] != nipypeVer)):
            pack2install.append('nipype.interfaces')

        # during the previous use of mia, mia_processes was not available or
        # its version was not known or its version was different from the one
        # currently available on the station
        if ((isinstance(proc_content, dict)) and ('Packages' in proc_content)
            and ('mia_processes' not in proc_content['Packages'])) or (
                (isinstance(proc_content, dict)) and ('Versions' in
                                                      proc_content)
                and ('mia_processes' not in proc_content['Versions'])) or (
                (isinstance(proc_content, dict)) and ('Versions' in
                                                      proc_content)
                and ('mia_processes' in proc_content['Versions'])
                and (proc_content['Versions']['mia_processes'] != miaProcVer)):
            pack2install.append('mia_processes')

    final_pckgs = dict()  # final_pckgs: the final dic of dic with the
    final_pckgs["Packages"] = {}  # informations about the
    final_pckgs["Versions"] = {}  # installed packages, their
    #             version, and the path to access
    #             them
    for pckg in pack2install:
        # pack2install: a list containing the package (nipype and/or
        # mia_processes) to install
        package = PackagesInstall()
        # pckg_dic: a dic of dic representation of a package and its
        # subpackages/modules Ex. {package: {subpackage: {pipeline:
        # 'process_enabled'}}}
        pckg_dic = package.add_package(pckg)

        for item in pckg_dic:
            final_pckgs["Packages"][item] = pckg_dic[item]

        if 'nipype' in pckg:  # Save the packages version
            final_pckgs["Versions"]["nipype"] = nipypeVer
            print('\n** Upgrading the {0} library processes to {1} version ...'
                  .format(pckg, nipypeVer))

        if 'mia_processes' in pckg:
            final_pckgs["Versions"]["mia_processes"] = miaProcVer
            print('\n** Upgrading the {0} library processes to {1} version ...'
                  .format(pckg, miaProcVer))

    if pack2install:

        if not any("nipype" in s for s in pack2install):
            print('\n** The nipype library processes in mia use already the '
                  'installed version ', nipypeVer)

        elif not any("mia_processes" in s for s in pack2install):
            print('\n** The mia_processes library in mia use already the '
                  'installed version ', miaProcVer)

        if (isinstance(proc_content, dict)) and ('Paths' in proc_content):

            for item in proc_content['Paths']:  # Save the path to the packages
                final_pckgs["Paths"] = proc_content['Paths']

        if (isinstance(proc_content, dict)) and ('Versions' in proc_content):

            for item in proc_content['Versions']:

                if item not in final_pckgs['Versions']:
                    final_pckgs['Versions'][item] = proc_content[
                        'Versions'][item]

        # try to keep the previous configuration before the update
        # of the packages
        if (isinstance(proc_content, dict)) and ('Packages' in proc_content):
            _deepCompDic(proc_content['Packages'], final_pckgs['Packages'])

            for item in proc_content['Packages']:

                if item not in final_pckgs['Packages']:
                    final_pckgs['Packages'][item] = proc_content[
                        'Packages'][item]

        with open(proc_config, 'w', encoding='utf8') as stream:
            yaml.dump(final_pckgs, stream, default_flow_style=False,
                      allow_unicode=True)

    else:
        print('\n- mia use already the installed version of nipype and '
              'mia_processes ({0} and {1} respectively)'
              .format(nipypeVer, miaProcVer))


if __name__ == '__main__':
    # this will only be executed when this module is run directly
    main()




# -*- coding: utf-8 -*- #

"""Module to handle the projects and their database."""

##########################################################################
# Populse_mia - Copyright (C) IRMaGe/CEA, 2018
# Distributed under the terms of the CeCILL license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL_V2.1-en.html
# for details.
##########################################################################



import sys

# Current version
version_major = 1
version_minor = 2
version_micro = 1
version_extra = ""

# Expected by setup.py: string of form "X.Y.Z"
__version__ = "{0}.{1}.{2}".format(version_major, version_minor, version_micro)


# Expected by setup.py: the status of the project
CLASSIFIERS = ['Development Status :: 5 - Production/Stable',
               'Intended Audience :: Developers',
               'License :: OSI Approved :: CEA CNRS Inria Logiciel Libre License, version 2.1 (CeCILL-2.1)',
               'Topic :: Software Development :: Libraries :: Python Modules',
               'Operating System :: OS Independent',
               'Programming Language :: Python :: 3.5',
               'Programming Language :: Python :: 3.6',
               'Programming Language :: Python :: 3.7',
               'Programming Language :: Python :: 3.8',
               'Programming Language :: Python :: 3 :: Only',
               'Topic :: Scientific/Engineering',
               'Topic :: Utilities']

# project descriptions
DESCRIPTION = 'populse mia'
LONG_DESCRIPTION = """
===============
populse_mia
===============
[MIA] Multiparametric Image Analysis:
A complete image processing environment mainly targeted at 
the analysis and visualization of large amounts of MRI data
"""

# Other values used in setup.py
NAME = 'populse_mia'
ORGANISATION = 'populse'
MAINTAINER = 'Populse team'
MAINTAINER_EMAIL = 'populse-support@univ-grenoble-alpes.fr'
AUTHOR = 'Populse team'
AUTHOR_EMAIL = 'populse-support@univ-grenoble-alpes.fr'
URL = 'http://populse.github.io/populse_mia'
DOWNLOAD_URL = 'http://populse.github.io/populse_mia'
LICENSE = 'CeCILL'
VERSION = __version__
CLASSIFIERS = CLASSIFIERS
PLATFORMS = 'OS Independent'

if sys.version_info < (3 , 6) and sys.version_info >= (3 , 5):
    REQUIRES = [
        'capsul',
        'lark-parser>=0.7.0',
        'matplotlib<3.1',
        'mia-processes>=1.2.1',
        'nibabel',
        'nipype',
        'pillow',
        'populse-db',
        'pyqt5',
        'python-dateutil',
        'pyyaml',
        'scikit-image',
        'scipy',
        'SIP',
        'sqlalchemy',
        'snakeviz',
        'soma_workflow',
        'traits',
]

elif sys.version_info >= (3 , 6):
    REQUIRES = [
        'capsul',
        'lark-parser>=0.7.0',
        'matplotlib',
        'mia-processes>=1.2.1',
        'nibabel',
        'nipype',
        'pillow',
        'populse-db',
        'pyqt5',
        'python-dateutil',
        'pyyaml',
        'scikit-image',
        'scipy',
        'SIP',  
        'sqlalchemy',
        'snakeviz',
        'soma_workflow',
        'traits',
]

else:
    # python < 3.5 is not compatible anyway
    REQUIRES = []

EXTRA_REQUIRES = {
    'doc': [
        'sphinx>=1.0',
    ],
}

brainvisa_build_model = 'pure_python'

# tests to run
test_commands = ['%s -m populse_mia.test' % sys.executable]



# -*- coding: utf-8 -*- #
"""

Contains:
    Class:
        - PipelineEditorTabs
        - PipelineEditor
    Function :
        - save_pipeline
        - get_path
        - find_filename

"""

##########################################################################
# Populse_mia - Copyright (C) IRMaGe/CEA, 2018
# Distributed under the terms of the CeCILL license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL_V2.1-en.html
# for details.
##########################################################################

import os
import sys
import six
import yaml
from traits.api import TraitError

# PyQt5 imports
from PyQt5 import QtGui, QtWidgets, QtCore

# Capsul imports
from capsul.api import get_process_instance, Process, PipelineNode, Switch
from capsul.qt_gui.widgets.pipeline_developper_view import \
    PipelineDevelopperView, NodeGWidget
from capsul.pipeline.xml import save_xml_pipeline
from capsul.pipeline.python_export import save_py_pipeline

# soma-base imports
from soma.utils.weak_proxy import weak_proxy

# Populse_MIA imports
from populse_mia.user_interface.pipeline_manager.node_controller import FilterWidget
from populse_mia.user_interface.pop_ups import PopUpClosePipeline
from populse_mia.software_properties import Config
from populse_mia.software_properties import verCmp

if sys.version_info[0] >= 3:
    unicode = str


    def values(d):
        return list(d.values())
else:
    def values(d):
        return d.values()


class PipelineEditorTabs(QtWidgets.QTabWidget):
    """
    Tab widget that contains pipeline editors

    Methods:
        - check_modifications: check if the nodes of the current pipeline have
           been modified
        - close_tab: close the selected tab and editor
        - emit_node_clicked: emit a signal when a node is clicked
        - emit_pipeline_saved: emit a signal when a pipeline is saved
        - emit_switch_clicked: emit a signal when a switch is clicked
        - export_to_db_scans: export the input of a filter to 'database_scans'
        - get_current_editor: get the instance of the current editor
        - get_current_filename: get the file name of the current pipeline
        - get_current_pipeline: get the instance of the current pipeline
        - get_current_tab_name: get the tab title of the current editor
        - get_filename_by_index: get the pipeline filename from its index in
           the editors
        - get_index_by_filename: get the index of the editor corresponding to
           the given pipeline filename
        - get_index_by_tab_name: get the index of the editor corresponding to
           the given tab name
        - get_tab_name_by_index: get the tab title from its index in the
           editors
        - has_pipeline_nodes: check if any of the pipelines in the editor tabs
           have pipeline nodes
        - load_pipeline: load a new pipeline
        - load_pipeline_parameters: load parameters to the pipeline of the
           current editor
        - new_tab: create a new tab and a new editor
        - open_filter: open a filter widget
        - open_sub_pipeline: open a sub-pipeline in a new tab
        - reset_pipeline: reset the pipeline of the current editor
        - save_pipeline: save the pipeline of the current editor
        - save_pipeline_parameters: save the pipeline parameters of the
           current editor
        - set_current_editor_by_tab_name: set the current editor
        - update_history: update undo/redo history of an editor
        - update_pipeline_editors: update editors
        - update_scans_list: update the list of database scans in every editor
    """

    pipeline_saved = QtCore.pyqtSignal(str)
    node_clicked = QtCore.pyqtSignal(str, Process)
    switch_clicked = QtCore.pyqtSignal(str, Switch)

    def __init__(self, project, scan_list, main_window):
        """Initialization of the Pipeline Editor tabs.

        :param project: current project in the software
        :param scan_list: list of the selected database files
        :param main_window: main window of the software
        """
        super(PipelineEditorTabs, self).__init__()

        self.project = project
        self.main_window = main_window
        self.setStyleSheet('QTabBar{font-size:12pt;font-family:Arial;'
                           'text-align: center;color:black;}')
        self.setTabsClosable(True)
        self.tabCloseRequested.connect(self.close_tab)
        self.scan_list = scan_list

        self.undos = {}
        self.redos = {}

        p_e = PipelineEditor(self.project, self.main_window)
        p_e.node_clicked.connect(self.emit_node_clicked)
        p_e.switch_clicked.connect(self.emit_switch_clicked)
        p_e.pipeline_saved.connect(self.emit_pipeline_saved)
        p_e.pipeline_modified.connect(self.update_pipeline_editors)
        p_e.edit_sub_pipeline.connect(self.open_sub_pipeline)
        p_e.open_filter.connect(self.open_filter)
        p_e.export_to_db_scans.connect(self.export_to_db_scans)

        # Setting a default editor called "New Pipeline"
        self.addTab(p_e, "New Pipeline")
        self.undos[p_e] = []
        self.redos[p_e] = []

        # Tool button to add a tab
        tb = QtWidgets.QToolButton()
        tb.setText('+')
        tb.clicked.connect(self.new_tab)

        self.addTab(QtWidgets.QLabel('Add tabs by pressing "+"'), str())
        self.setTabEnabled(1, False)
        self.tabBar().setTabButton(1, QtWidgets.QTabBar.RightSide, tb)

        # Checking if the pipeline nodes have been modified
        self.tabBarClicked.connect(self.check_modifications)

    def check_modifications(self, current_index):
        """Check if the nodes of the current pipeline have been modified"""

        # If the user click on the last tab (with the '+'),
        # it will throw an AttributeError
        try:
            self.widget(current_index).check_modifications()
        except AttributeError:
            pass

    def close_tab(self, idx):
        """Close the selected tab and editor.

        :param idx: index of the tab to close
        """

        filename = os.path.basename(self.get_filename_by_index(idx))
        editor = self.get_editor_by_index(idx)

        # If the pipeline has been modified and not saved
        if self.tabText(idx)[-2:] == " *":
            self.pop_up_close = PopUpClosePipeline(filename)
            self.pop_up_close.save_as_signal.connect(self.save_pipeline)
            self.pop_up_close.exec()

            can_exit = self.pop_up_close.can_exit()

            if self.pop_up_close.bool_save_as:
                if idx == self.currentIndex():
                    self.setCurrentIndex(max(0, self.currentIndex() - 1))
                self.removeTab(idx)
                return

        else:
            can_exit = True

        if not can_exit:
            return

        del self.undos[editor]
        del self.redos[editor]

        if idx == self.currentIndex():
            self.setCurrentIndex(max(0, self.currentIndex() - 1))
        self.removeTab(idx)

        # If there is no more editor, adding one
        if self.count() == 1:
            p_e = PipelineEditor(self.project, self.main_window)
            p_e.node_clicked.connect(self.emit_node_clicked)
            p_e.switch_clicked.connect(self.emit_switch_clicked)
            p_e.pipeline_saved.connect(self.emit_pipeline_saved)
            p_e.pipeline_modified.connect(self.update_pipeline_editors)
            p_e.edit_sub_pipeline.connect(self.open_sub_pipeline)
            p_e.open_filter.connect(self.open_filter)
            p_e.export_to_db_scans.connect(self.export_to_db_scans)

            # Setting a default editor called "New Pipeline"
            self.insertTab(0, p_e, "New Pipeline")
            self.setCurrentIndex(0)
            self.undos[p_e] = []
            self.redos[p_e] = []

    def emit_node_clicked(self, node_name, process):
        """Emit a signal when a node is clicked.

        :param node_name: node name
        :param process: process of the corresponding node
        """

        self.node_clicked.emit(node_name, process)

    def emit_pipeline_saved(self, filename):
        """Emit a signal when a pipeline is saved.

        :param filename: file name of the pipeline
        """

        self.setTabText(self.currentIndex(), os.path.basename(filename))
        self.pipeline_saved.emit(filename)

    def emit_switch_clicked(self, node_name, switch):
        """Emit a signal when a switch is clicked.

        :param node_name: node name
        :param switch: process of the corresponding node
        """

        self.switch_clicked.emit(node_name, switch)

    def export_to_db_scans(self, node_name):
        """Export the input of a filter to "database_scans" plug.

        :param node_name:
        """

        # If database_scans is already a pipeline global input, the plug
        # cannot be exported. A link as to be added between database_scans
        # and the input of the filter.
        if 'database_scans' in \
                self.get_current_pipeline().user_traits().keys():
            self.get_current_pipeline().add_link('database_scans->{0}'
                                                 '.input'.format(node_name))
        else:
            self.get_current_pipeline().export_parameter(
                node_name, 'input',
                pipeline_parameter='database_scans')
        self.get_current_editor().scene.update_pipeline()

    def get_current_editor(self):
        """Get the instance of the current editor.

        :return: the current editor
        """

        return self.get_editor_by_index(self.currentIndex())

    def get_current_filename(self):
        """Get the relative path to the file the pipeline in the current editor
           has been last saved to.
        If the pipeline has never been saved, returns the title of the tab.

        :return: the filename of the current editor
        """

        return self.get_filename_by_index(self.currentIndex())

    def get_current_pipeline(self):
        """Get the instance of the current pipeline.

        :return: the pipeline of the current editor
        """

        return self.get_current_editor().scene.pipeline

    def get_current_tab_name(self):
        """Get the tab name of the editor in the current tab.
        Trailing " \*" and ampersand ("&") characters are removed.

        :return: the current tab name
        """

        return self.get_tab_name_by_index(self.currentIndex())

    def get_editor_by_file_name(self, file_name):
        """Get the instance of an editor from its file name.

        :param file_name: name of the file the pipeline was last saved to
        :return: the editor corresponding to the file name
        """

        return self.get_editor_by_index(self.get_index_by_filename(file_name))

    def get_editor_by_index(self, idx):
        """Get the instance of an editor from its index in the editors.

        :param idx: index of the editor
        :return: the editor corresponding to the index
        """

        # last tab has "add tab" button, no editor
        if idx in range(self.count() - 1):
            return self.widget(idx)

    def get_editor_by_tab_name(self, tab_name):
        """Get the instance of an editor from its tab name.

        :param tab_name: name of the tab
        :return: the editor corresponding to the tab name
        """

        return self.get_editor_by_index(self.get_index_by_tab_name(tab_name))

    def get_filename_by_index(self, idx):
        """Get the relative path to the file the pipeline in the editor at the
           given index has been last saved to.
        If the pipeline has never been saved, returns the title of the tab.

        :param idx: index of the editor
        :return: the file name corresponding to the index
        """

        editor = self.get_editor_by_index(idx)
        if editor is not None:
            return editor.get_current_filename()

    def get_index_by_editor(self, editor):
        """Get the index of the editor corresponding to the given editor.

        :param editor: searched pipeline editor
        :return: the index corresponding to the editor
        """
        for idx in range(self.count() - 1):
            if self.get_editor_by_index(idx) == editor:
                return idx

    def get_index_by_filename(self, filename):
        """Get the index of the first editor corresponding to the given
           pipeline filename

        :param filename: filename of the searched pipeline
        :return: the index corresponding to the file name
        """

        if filename:
            # we always store file names as relative paths
            filename = os.path.relpath(filename)

            for idx in range(self.count() - 1):
                if self.get_filename_by_index(idx) == filename:
                    return idx

    def get_index_by_tab_name(self, tab_name):
        """Get the index of the editor corresponding to the given tab name.

        :param tab_name: name of the tab with the searched pipeline
        :return: the index corresponding to the tab name
        """

        for idx in range(self.count() - 1):
            if self.get_tab_name_by_index(idx) == tab_name:
                return idx

    def get_tab_name_by_index(self, idx):
        """Get the tab name of the editor at the given index.
        Trailing " \*" and ampersand ("&") characters are removed.

        :param idx: index of the editor
        :return: the tab name corresponding to the index
        """

        # last tab has "add tab" button, no tab name
        if idx in range(self.count() - 1):
            # remove Qt keyboard shortcut indicator
            tab_name = self.tabText(idx).replace("&", "", 1)
            if tab_name[-2:] == " *":
                tab_name = tab_name[:-2]

            return tab_name

    def has_pipeline_nodes(self):
        """Check if any of the pipelines in the editor tabs have pipeline nodes

        :return: True or False depending on if there are nodes in the editors
        """
        for idx in range(self.count()):
            p_e = self.widget(idx)
            if hasattr(p_e, 'scene'):
                # if the widget is a tab editor
                if p_e.scene.pipeline.nodes[''].plugs:
                    return True

        return False

    def load_pipeline(self, filename=None):
        """Load a new pipeline.

        :param filename: not None only when this method is called from
          "open_sub_pipeline"
        """

        current_tab_not_empty = len(
            self.get_current_editor().scene.pipeline.nodes.keys()) > 1
        new_tab_opened = False

        if filename is None:
            # Open new tab if the current PipelineEditor is not empty
            if current_tab_not_empty:
                # create new tab with new editor and make it current
                self.new_tab()
                working_index = self.currentIndex()
                new_tab_opened = True
            # get only the file name to load
            filename = self.get_current_editor().load_pipeline('', False)

        if filename:
            # Check if this pipeline is already open
            existing_pipeline_tab = self.get_index_by_filename(filename)

            if existing_pipeline_tab is not None:
                self.setCurrentIndex(existing_pipeline_tab)
            else:  # we need to actually load the pipeline
                if current_tab_not_empty and not new_tab_opened:
                    self.new_tab()
                    new_tab_opened = True

                working_index = self.currentIndex()
                editor = self.get_editor_by_index(working_index)
                # actually load the pipeline
                filename = editor.load_pipeline(filename)
                if filename:
                    self.setTabText(working_index, os.path.basename(filename))
                    self.update_scans_list()
                    return  # success

        # if we're still here, something went wrong. clean up.
        if new_tab_opened:
            self.close_tab(working_index)

    def load_pipeline_parameters(self):
        """Load parameters to the pipeline of the current editor"""

        self.get_current_editor().load_pipeline_parameters()

    def new_tab(self):
        """Create a new tab and a new editor and makes the new tab current."""

        # Creating a new editor
        p_e = PipelineEditor(self.project, self.main_window)
        p_e.node_clicked.connect(self.emit_node_clicked)
        p_e.switch_clicked.connect(self.emit_switch_clicked)
        p_e.pipeline_saved.connect(self.emit_pipeline_saved)
        p_e.pipeline_modified.connect(self.update_pipeline_editors)
        p_e.edit_sub_pipeline.connect(self.open_sub_pipeline)
        p_e.open_filter.connect(self.open_filter)
        p_e.export_to_db_scans.connect(self.export_to_db_scans)

        # A unique editor name has to be automatically generated
        idx = 1
        while True and idx < 50:
            name = "New Pipeline {0}".format(idx)
            if self.get_index_by_tab_name(name):
                idx += 1
                continue
            else:
                break
        if name is not None:
            self.undos[p_e] = []
            self.redos[p_e] = []
        else:
            print('Too many tabs in the Pipeline Editor')
            return

        self.insertTab(self.count() - 1, p_e, name)
        self.setCurrentIndex(self.count() - 2)

    def open_filter(self, node_name):
        """Open a filter widget.

        :param node_name: name of the corresponding node
        """
        node = self.get_current_pipeline().nodes[node_name]
        self.filter_widget = FilterWidget(self.project, node_name, node, self)
        self.filter_widget.show()

    def open_sub_pipeline(self, sub_pipeline):
        """Open a sub-pipeline in a new tab.

        :param sub_pipeline: the pipeline to open
        """

        # Reading the process configuration file
        config = Config()
        with open(os.path.join(config.get_mia_path(),
                               'properties',
                               'process_config.yml'), 'r') as stream:

            try:
                if verCmp(yaml.__version__, '5.1', 'sup'):
                    dic = yaml.load(stream, Loader=yaml.FullLoader)

                else:
                    dic = yaml.load(stream)

            except yaml.YAMLError as exc:
                print(exc)
                dic = {}

        sub_pipeline_name = sub_pipeline.name

        # get_path returns a list that is the package path to
        # the sub_pipeline file
        sub_pipeline_list = get_path(sub_pipeline_name, dic['Packages'])
        sub_pipeline_name = sub_pipeline_list.pop()

        # Finding the real sub-pipeline filename
        sub_pipeline_filename = find_filename(
            dic['Paths'], sub_pipeline_list, sub_pipeline_name)
        self.load_pipeline(sub_pipeline_filename)

    def reset_pipeline(self):
        """Reset the pipeline of the current editor"""

        self.get_current_editor()._reset_pipeline()

    def save_pipeline(self, new_file_name=None):
        """Save the pipeline of the current editor."""
        if new_file_name is None:
            # Doing a "Save as" action
            new_file_name = os.path.basename(
                self.get_current_editor().save_pipeline())

            if (new_file_name and
                    os.path.basename(
                        self.get_current_filename()) != new_file_name):
                self.setTabText(self.currentIndex(), new_file_name)
        else:
            # Saving the current pipeline
            pipeline = self.get_current_pipeline()

            try:
                posdict = dict(
                    [(key, (value.x(), value.y())) for key, value in
                     six.iteritems(self.get_current_editor().scene.pos)])

            except:  # add by Irmage OM
                posdict = dict(
                    [(key, (value[0], value[1])) for key, value in
                     six.iteritems(self.get_current_editor().scene.pos)])

            # add by Irmage OM
            dimdict = dict(
                [(key, (value[0], value[1])) for key, value in
                 six.iteritems(self.get_current_editor().scene.dim)])

            # add by Irmage OM
            pipeline.node_dimension = dimdict

            old_pos = pipeline.node_position
            pipeline.node_position = posdict
            save_pipeline(pipeline, new_file_name)
            self.get_current_editor()._pipeline_filename = unicode(
                new_file_name)
            pipeline.node_position = old_pos
            self.pipeline_saved.emit(new_file_name)
            self.setTabText(self.currentIndex(),
                            os.path.basename(new_file_name))

    def save_pipeline_parameters(self):
        """Save the pipeline parameters of the current editor"""

        self.get_current_editor().save_pipeline_parameters()

    def set_current_editor_by_editor(self, editor):
        """Set the current editor.

        :param editor: editor in the tab that should be made current
        """

        self.setCurrentIndex(self.get_index_by_editor(editor))

    def set_current_editor_by_file_name(self, file_name):
        """Set the current editor.

        :param file_name: name of the file the pipeline was last saved to
        """

        self.setCurrentIndex(self.get_index_by_filename(file_name))

    def set_current_editor_by_tab_name(self, tab_name):
        """Set the current editor.

        :param tab_name: name of the tab
        """

        self.setCurrentIndex(self.get_index_by_tab_name(tab_name))

    def update_history(self, editor):
        """Update undo/redo history of an editor

        :param editor: editor
        """

        self.undos[editor] = editor.undos
        self.redos[editor] = editor.redos
        self.setTabText(self.currentIndex(),
                        self.get_current_tab_name() + " *")
        # make sure the " *" is there

    def update_pipeline_editors(self, editor):
        """Update editor.

        :param editor: editor
        """

        self.update_history(editor)
        self.update_scans_list()

    def update_scans_list(self):
        """Update the list of database scans in every editor"""

        for i in range(self.count() - 1):
            pipeline = self.widget(i).scene.pipeline
            if hasattr(pipeline, "nodes"):
                for node_name, node in pipeline.nodes.items():
                    if node_name == "":
                        for plug_name, plug in node.plugs.items():
                            if plug_name == "database_scans":
                                node.set_plug_value(plug_name, self.scan_list)


class PipelineEditor(PipelineDevelopperView):
    """View to edit a pipeline graphically

    Methods:
        - _del_link: deletes a link
        - _export_plug: export a plug to a pipeline global input or output
        - _release_grab_link: method called when a link is released
        - _remove_plug: removes a plug
        - add_link: add a link between two nodes
        - add_process: adds a process to the pipeline
        - check_modifications: checks if the nodes of the pipeline have been
           modified
        - del_node: deletes a node
        - dragEnterEvent: event handler when the mouse enters the widget
        - dragMoveEvent: event handler when the mouse moves in the widget
        - dropEvent: event handler when something is dropped in the widget
        - export_node_plugs: exports all the plugs of a node
        - find_process: finds the dropped process in the system's paths
        - get_current_filename: returns the relative path the pipeline was
           last saved to. Empty if never saved.
        - save_pipeline: saves the pipeline
        - update_history: updates the history for undos and redos
        - update_node_name: updates a node name
        - update_plug_value: updates a plug value
    """

    pipeline_saved = QtCore.pyqtSignal(str)
    pipeline_modified = QtCore.pyqtSignal(PipelineDevelopperView)

    def __init__(self, project, main_window):
        """Initialization of the PipelineEditor.

        :param project: current project in the software
        """

        PipelineDevelopperView.__init__(self, pipeline=None,
                                        allow_open_controller=True,
                                        show_sub_pipelines=True,
                                        enable_edition=True)

        self.project = project
        self.main_window = main_window

        # Undo/Redo
        self.undos = []
        self.redos = []

    def _del_link(self, link=None, from_undo=False, from_redo=False):
        """Delete a link.

        :param link: string representation of a link
           (e.g. "process1.out->process2.in")
        :param from_undo: boolean, True if the action has been made using an
           undo
        :param from_redo: boolean, True if the action has been made using a
           redo
        """

        if not link:
            link = self._current_link
        else:
            self._current_link = link

        (source_node_name, source_plug_name, source_node,
         source_plug, dest_node_name, dest_plug_name, dest_node,
         dest_plug) = self.scene.pipeline.parse_link(link)

        (dest_node_name, dest_parameter, dest_node, dest_plug,
         weak_link) = list(source_plug.links_to)[0]

        active = source_plug.activated and dest_plug.activated

        # Calling the original method
        PipelineDevelopperView._del_link(self)

        # For history
        history_maker = ["delete_link", (source_node_name, source_plug_name),
                         (dest_node_name, dest_plug_name), active, weak_link]

        self.update_history(history_maker, from_undo, from_redo)

        self.main_window.statusBar().showMessage(
            'Link {0} has been deleted.'.format(link))

    def _export_plug(self, pipeline_parameter=False, optional=None,
                     weak_link=None, from_undo=False, from_redo=False,
                     temp_plug_name=None):
        """Export a plug to a pipeline global input or output.

        :param pipeline_parameter: name of the pipeline input/output
        :param optional: True if the plug is optional
        :param weak_link: True if the link is weak
        :param from_undo: True if this method is called from an undo action
        :param from_redo: True if this method is called from a redo action
        :param temp_plug_name: tuple containing (the name of the node, the
           name of the plug) to export
        """
        # Bug: the first parameter (here pipeline_parameter) cannot be None
        # even if we write pipeline_parameter=None in the line above,
        # it will be False...

        if temp_plug_name is None:
            dial = self._PlugEdit()
            dial.name_line.setText(self._temp_plug_name[1])
            dial.optional.setChecked(self._temp_plug.optional)
            temp_plug_name = self._temp_plug_name

            res = dial.exec_()
        else:
            res = True

        if res:
            if not pipeline_parameter:
                pipeline_parameter = str(dial.name_line.text())

            if optional is None:
                optional = dial.optional.isChecked()

            if weak_link is None:
                weak_link = dial.weak.isChecked()

            try:
                self.scene.pipeline.export_parameter(
                    temp_plug_name[0], temp_plug_name[1],
                    pipeline_parameter=pipeline_parameter,
                    is_optional=optional,
                    weak_link=weak_link)
            except TraitError:
                print("Cannot export {0}.{1} plug".format(temp_plug_name[0],
                                                          temp_plug_name[1]))

            '''self.scene.pipeline.export_parameter(
                temp_plug_name[0], temp_plug_name[1],
                pipeline_parameter=pipeline_parameter,
                is_optional=optional,
                weak_link=weak_link)'''  # Uncomment to generate the error

            self.scene.update_pipeline()

        # For history
        history_maker = ["export_plug", ('inputs', pipeline_parameter),
                         pipeline_parameter, optional, weak_link]

        self.update_history(history_maker, from_undo, from_redo)

        self.main_window.statusBar().showMessage(
            "Plug {0} has been exported.".format(temp_plug_name[1]))

    def _release_grab_link(self, event, ret=False):
        """Method called when a link is released.

        :param event: mouse event corresponding to the release
        :param ret: boolean that is set to True in the original method to
           return the link
        """

        # Calling the original method
        link = PipelineDevelopperView._release_grab_link(self, event, ret=True)

        # For history
        history_maker = ["add_link", link]

        self.update_history(history_maker, from_undo=False, from_redo=False)

        self.main_window.statusBar().showMessage(
            'Link {0} has been added.'.format(link))

    def add_link(self, source, dest, active, weak,
                 from_undo=False, from_redo=False):
        """Add a link between two nodes.

        :param source: tuple containing the node and plug source names
        :param dest: tuple containing the node and plug destination names
        :param active: boolean that is True if the link is activated
        :param weak: boolean that is True if the link is weak
        :param from_undo: boolean that is True if the action has been made
           using an undo
        :param from_redo: boolean that is True if the action has been made
           using a redo
        """

        self.scene.add_link(source, dest, active, weak)

        # Writing a string to represent the link
        source_parameters = ".".join(source)
        dest_parameters = ".".join(dest)
        link = "->".join((source_parameters, dest_parameters))

        self.scene.pipeline.add_link(link)
        self.scene.update_pipeline()

        # For history
        history_maker = ["add_link", link]

        self.update_history(history_maker, from_undo, from_redo)

        self.main_window.statusBar().showMessage(
            'Link {0} has been added.'.format(link))

    def add_process(self, class_process, node_name=None,
                    from_undo=False, from_redo=False, links=[]):
        """Add a process to the pipeline.

        :param class_process: process class's name (str)
        :param node_name: name of the corresponding node (using when undo/redo) (str)
        :param from_undo: boolean that is True if the action has been made using an undo
        :param from_redo: boolean that is True if the action has been made using a redo
        :param links: list of links (using when undo/redo)
        """

        pipeline = self.scene.pipeline
        if not node_name:
            class_name = class_process.__name__
            i = 1

            node_name = class_name.lower() + str(i)

            while node_name in pipeline.nodes and i < 100:
                i += 1
                node_name = class_name.lower() + str(i)

            process_to_use = class_process()

        else:
            process_to_use = class_process

        try:
            process = get_process_instance(
                process_to_use)
        except Exception as e:
            print(e)
            return

        if hasattr(process, 'use_project') and process.use_project:
            process.project = self.project

        pipeline.add_process(node_name, process)

        # Capsul update
        node = pipeline.nodes[node_name]
        gnode = self.scene.add_node(node_name, node)
        gnode.setPos(self.mapToScene(self.mapFromGlobal(self.click_pos)))

        # If the process is added from a undo, all the links
        # that were connected to the corresponding node has to be reset
        for link in links:
            source = link[0]
            dest = link[1]
            active = link[2]
            weak = link[3]
            self.scene.add_link(source, dest, active, weak)
            # Writing a string to represent the link
            source_parameters = ".".join(source)
            dest_parameters = ".".join(dest)
            link_to_add = "->".join((source_parameters, dest_parameters))
            self.scene.pipeline.add_link(link_to_add)
            self.scene.update_pipeline()

        # For history
        history_maker = ["add_process"]

        if from_undo:
            # Adding the arguments to make the redo correctly
            history_maker.append(node_name)
        else:
            # Adding the arguments to make the undo correctly
            history_maker.append(node_name)
            history_maker.append(class_process)

        self.update_history(history_maker, from_undo, from_redo)

        self.main_window.statusBar().showMessage(
            "Node {0} has been added.".format(node_name))

    def check_modifications(self):
        """Check if the nodes of the pipeline have been modified."""

        pipeline = self.scene.pipeline
        config = Config()

        # List to store the removed links
        removed_links = []

        for node_name, node in pipeline.nodes.items():
            # Only if the node is a pipeline node ?
            if node_name and isinstance(node, PipelineNode):
                sub_pipeline_process = node.process
                current_process_id = sub_pipeline_process.id

                # Reading the process configuration file
                with open(os.path.join(config.get_mia_path(),
                                       'properties',
                                       'process_config.yml'), 'r') as stream:

                    try:
                        if verCmp(yaml.__version__, '5.1', 'sup'):
                            dic = yaml.load(stream,
                                            Loader=yaml.FullLoader)
                        else:
                            dic = yaml.load(stream)

                    except yaml.YAMLError as exc:
                        print(exc)
                        dic = {}

                sub_pipeline_name = sub_pipeline_process.name

                # get_path returns a list that is the package path
                # to the sub_pipeline file
                sub_pipeline_list = get_path(sub_pipeline_name,
                                             dic['Packages'])
                sub_pipeline_name = sub_pipeline_list.pop()

                # Finding the real sub-pipeline filename
                sub_pipeline_filename = find_filename(
                    dic['Paths'], sub_pipeline_list, sub_pipeline_name)

                saved_process = get_process_instance(sub_pipeline_filename)

                current_inputs = list(sub_pipeline_process.get_inputs().keys())
                current_outputs = list(
                    sub_pipeline_process.get_outputs().keys())
                saved_inputs = list(saved_process.get_inputs().keys())
                saved_outputs = list(saved_process.get_outputs().keys())

                new_inputs = [item for item in saved_inputs
                              if item not in current_inputs]
                new_outputs = [item for item in saved_outputs
                               if item not in current_outputs]
                removed_inputs = [item for item in current_inputs
                                  if item not in saved_inputs]
                removed_outputs = [item for item in current_outputs
                                   if item not in saved_outputs]

                if new_inputs or new_outputs or removed_inputs or \
                        removed_outputs:

                    # Checking the links of the node
                    link_to_del = set()
                    for link, glink in six.iteritems(self.scene.glinks):
                        if link[0][0] == node_name or link[1][0] == node_name:
                            self.scene.removeItem(glink)
                            link_to_del.add((link, glink))
                    for link, glink in link_to_del:
                        del self.scene.glinks[link]

                    # Creating a new node
                    new_node = saved_process.pipeline_node
                    new_node.name = node_name
                    new_node.pipeline = pipeline
                    saved_process.parent_pipeline = weak_proxy(pipeline)
                    pipeline.nodes[node_name] = new_node

                    # Creating a new graphical node
                    gnode = NodeGWidget(
                        node_name, new_node.plugs, pipeline,
                        sub_pipeline=saved_process, process=saved_process,
                        colored_parameters=self.scene.colored_parameters,
                        logical_view=self.scene.logical_view,
                        labels=self.scene.labels)

                    # Setting the new node to the same position as the
                    # previous one
                    pos = self.scene.pos.get(node_name)
                    gnode.setPos(pos)

                    # Removing the old node
                    self.scene.removeItem(self.scene.gnodes[node_name])
                    del self.scene.gnodes[node_name]

                    # Adding the new node to the scene
                    self.scene.gnodes[node_name] = gnode
                    self.scene.addItem(gnode)

                    for link, glink in link_to_del:
                        source = link[0]
                        dest = link[1]
                        if source[0] == 'inputs':
                            link_plug = source[1]
                            source = ('', link_plug)
                        if dest[0] == 'outputs':
                            link_plug = dest[1]
                            dest = ('', link_plug)

                        # Writing a string to represent the link
                        source_parameters = ".".join(source)
                        dest_parameters = ".".join(dest)
                        pipeline_link = "->".join((source_parameters,
                                                   dest_parameters))

                        # Checking if a connected plug has been deleting
                        if dest[0] == node_name and dest[1] in removed_inputs:
                            removed_links.append(pipeline_link)
                            continue

                        if source[0] == node_name and source[1] in \
                                removed_outputs:
                            removed_links.append(pipeline_link)
                            continue

                        # Adding the link
                        self.scene.pipeline.add_link(pipeline_link)
                        self.scene.add_link(source, dest, True, False)

                    self.scene.update_pipeline()

        if removed_links:
            dialog_text = 'Pipeline {0} has been updated.\n' \
                          'Removed links:'.format(node_name)
            for removed_link in removed_links:
                dialog_text += '\n{0}'.format(removed_link)

            pop_up = QtWidgets.QMessageBox()
            pop_up.setIcon(QtWidgets.QMessageBox.Warning)
            pop_up.setText(dialog_text)
            pop_up.setWindowTitle("Warning: links have been removed")
            pop_up.setStandardButtons(
                QtWidgets.QMessageBox.Ok | QtWidgets.QMessageBox.Cancel)
            pop_up.exec_()

    def get_current_filename(self):
        """Return the relative path the pipeline was last saved to.
        Empty if never saved.

        :return: the current pipeline file name
        """
        if hasattr(self, '_pipeline_filename') and self._pipeline_filename:
            return os.path.relpath(self._pipeline_filename)
        else:
            return ''

    def del_node(self, node_name=None, from_undo=False, from_redo=False):
        """Delete a node.

        :param node_name: name of the corresponding node (using when undo/redo)
        :param from_undo: boolean, True if the action has been made using an
           undo
        :param from_redo: boolean, True if the action has been made using a
           redo
        """

        pipeline = self.scene.pipeline
        if not node_name:
            node_name = self.current_node_name
        node = pipeline.nodes[node_name]

        # Collecting the links from the node that is being deleted
        links = []
        for plug_name, plug in node.plugs.items():
            if plug.output:
                for link_to in plug.links_to:
                    (dest_node_name, dest_parameter, dest_node, dest_plug,
                     weak_link) = link_to
                    active = plug.activated

                    # Looking for the name of dest_plug in dest_node
                    dest_plug_name = None
                    for plug_name_d, plug_d in dest_node.plugs.items():
                        if plug_d == dest_plug:
                            dest_plug_name = plug_name_d
                            break

                    link_to_add = [(node_name, plug_name)]
                    link_to_add.append((dest_node_name, dest_plug_name))
                    link_to_add.append(active)
                    link_to_add.append(weak_link)

                    links.append(link_to_add)
            else:
                for link_from in plug.links_from:
                    (source_node_name, source_parameter, source_node,
                     source_plug, weak_link) = link_from
                    active = plug.activated

                    # Looking for the name of source_plug in source_node
                    source_plug_name = None
                    for plug_name_d, plug_d in source_node.plugs.items():
                        if plug_d == source_plug:
                            source_plug_name = plug_name_d
                            break

                    link_to_add = [(source_node_name, source_plug_name)]
                    link_to_add.append((node_name, plug_name))
                    link_to_add.append(active)
                    link_to_add.append(weak_link)

                    links.append(link_to_add)

        # Calling the original method
        PipelineDevelopperView.del_node(self, node_name)

        # For history
        history_maker = ["delete_process", node_name, node.process, links]

        self.update_history(history_maker, from_undo, from_redo)

        self.main_window.statusBar().showMessage(
            "Node {0} has been deleted.".format(node_name))

    def dragEnterEvent(self, event):
        """Event handler when the mouse enters the widget.

        :param event: event
        """

        if event.mimeData().hasFormat('component/name'):
            event.accept()

    def dragMoveEvent(self, event):
        """Event handler when the mouse moves in the widget.

        :param event: event
        """

        if event.mimeData().hasFormat('component/name'):
            event.accept()

    def dropEvent(self, event):
        """Event handler when something is dropped in the widget.

        :param event: event

        """

        if event.mimeData().hasFormat('component/name'):
            self.click_pos = QtGui.QCursor.pos()
            path = bytes(event.mimeData().data('component/name'))
            self.find_process(path.decode('utf8'))

    def find_process(self, path):
        """Find the dropped process in the system's paths.

        :param path: class's path (e.g. "nipype.interfaces.spm.Smooth") (str)
        """
        package_name, process_name = os.path.splitext(path)
        process_name = process_name[1:]
        __import__(package_name)
        pkg = sys.modules[package_name]
        for name, instance in sorted(list(pkg.__dict__.items())):
            if name == process_name:
                try:
                    process = get_process_instance(instance)
                except Exception as e:
                    print(e)
                    return
                else:
                    QtGui.QApplication.setOverrideCursor(QtCore.Qt.WaitCursor)
                    self.add_process(instance)
                    QtGui.QApplication.restoreOverrideCursor()

    def update_history(self, history_maker, from_undo, from_redo):
        """Update the history for undos and redos.
        This method is called after each action in the PipelineEditor.

        :param history_maker: list that contains information about what has
            been done
        :param from_undo: boolean that is True if the action has been made
            using an undo
        :param from_redo: boolean that is True if the action has been made
            using a redo
        """

        if from_undo:
            self.redos.append(history_maker)
        else:
            self.undos.append(history_maker)
            # If the action does not come from an undo or a redo,
            # the redos has to be cleared
            if not from_redo:
                self.redos.clear()

        self.pipeline_modified.emit(self)

    def update_node_name(self, old_node, old_node_name, new_node_name,
                         from_undo=False, from_redo=False):
        """Update a node name.

        :param old_node: Node object to change
        :param old_node_name: original name of the node (str)
        :param new_node_name: new name of the node (str)
        :param from_undo: boolean, True if the action has been made using an
           undo
        :param from_redo: boolean, True if the action has been made using a
           redo
        """

        pipeline = self.scene.pipeline

        # Removing links of the selected node and copy the origin/destination
        links_to_copy = []
        for source_parameter, source_plug \
                in six.iteritems(old_node.plugs):
            for (dest_node_name, dest_parameter, dest_node, dest_plug,
                 weak_link) in source_plug.links_to.copy():
                pipeline.remove_link(old_node_name + "." + source_parameter +
                                     "->" + dest_node_name + "." +
                                     dest_parameter)
                links_to_copy.append(("to", source_parameter,
                                      dest_node_name, dest_parameter))

            for (dest_node_name, dest_parameter, dest_node, dest_plug,
                 weak_link) in source_plug.links_from.copy():
                pipeline.remove_link(dest_node_name + "." + dest_parameter +
                                     "->" + old_node_name + "." +
                                     source_parameter)
                links_to_copy.append(("from", source_parameter, dest_node_name,
                                      dest_parameter))

        # Creating a new node with the new name and deleting the previous one
        pipeline.nodes[new_node_name] = pipeline.nodes[old_node_name]
        del pipeline.nodes[old_node_name]

        # Setting the same links as the original node
        for link in links_to_copy:

            if link[0] == "to":
                pipeline.add_link(new_node_name + "." + link[1] + "->"
                                  + link[2] + "." + link[3])
            elif link[0] == "from":
                pipeline.add_link(link[2] + "." + link[3] + "->"
                                  + new_node_name + "." + link[1])

        # Updating the pipeline
        pipeline.update_nodes_and_plugs_activation()

        # For history
        history_maker = ["update_node_name", pipeline.nodes[new_node_name],
                         new_node_name, old_node_name]

        self.update_history(history_maker, from_undo, from_redo)

        self.main_window.statusBar().showMessage(
            'Node name "{0}" has been changed to "{1}".'.format(
                old_node_name, new_node_name))

    def update_plug_value(self, node_name, new_value, plug_name, value_type,
                          from_undo=False, from_redo=False):
        """Update a plug value.

        :param node_name: name of the node (str)
        :param new_value: new value to set to the plug
        :param plug_name: name of the plug to change (str)
        :param value_type: type of the new value
        :param from_undo: boolean, True if the action has been made using an
        undo
        :param from_redo: boolean, True if the action has been made using a
        redo
        """

        old_value = self.scene.pipeline.nodes[node_name].get_plug_value(
            plug_name)
        self.scene.pipeline.nodes[node_name].set_plug_value(
            plug_name, value_type(new_value))

        # For history
        history_maker = ["update_plug_value", node_name, old_value,
                         plug_name, value_type]

        self.update_history(history_maker, from_undo, from_redo)

        self.main_window.statusBar().showMessage(
            'Plug "{0}" of node "{1}" has been changed to "{2}".'.format(
                plug_name, node_name, new_value))

    def export_node_plugs(self, node_name, inputs=True, outputs=True,
                          optional=False, from_undo=False, from_redo=False):
        """Export all the plugs of a node

        :param node_name: node name
        :param inputs: True if the inputs have to be exported
        :param outputs: True if the outputs have to be exported
        :param optional: True if the optional plugs have to be exported
        :param from_undo: True if this method is called from an undo action
        :param from_redo: True if this method is called from a redo action
        """

        pipeline = self.scene.pipeline
        node = pipeline.nodes[node_name]

        parameter_list = []
        for parameter_name, plug in six.iteritems(node.plugs):
            if parameter_name in ("nodes_activation", "selection_changed"):
                continue
            if (((node_name, parameter_name) not in pipeline.do_not_export and
                 ((outputs and plug.output and not plug.links_to) or
                  (inputs and not plug.output and not plug.links_from)) and
                 (optional or not node.get_trait(parameter_name).optional))):
                try:
                    pipeline.export_parameter(node_name, parameter_name)
                    parameter_list.append(parameter_name)
                except TraitError:
                    print("Cannot export {0}.{1} plug".format(node_name,
                                                              parameter_name))

        # For history
        history_maker = ["export_plugs", parameter_list, node_name]

        self.update_history(history_maker, from_undo, from_redo)

        self.main_window.statusBar().showMessage(
            "Plugs {0} have been exported.".format(str(parameter_list)))

    def _remove_plug(self, _temp_plug_name=None, from_undo=False,
                     from_redo=False, from_export_plugs=False):
        """Remove a plug

        :param _temp_plug_name: tuple containing (the name of the node,
           the name of the plug) to remove
        :param from_undo: True if this method is called from an undo action
        :param from_redo: True if this method is called from a redo action
        :param from_export_plugs: True if this method is called from a
           "from_export_plugs" undo or redo action
        """
        if not _temp_plug_name:
            _temp_plug_name = self._temp_plug_name

        if _temp_plug_name[0] in ('inputs', 'outputs'):
            plug_name = _temp_plug_name[1]
            plug = self.scene.pipeline.pipeline_node.plugs[plug_name]
            optional = plug.optional
            # contains the plugs that are connected to the input or output plug
            new_temp_plugs = []
            for link in plug.links_to:
                temp_plug = (link[0], link[1])
                new_temp_plugs.append(temp_plug)
            for link in plug.links_from:
                temp_plug = (link[0], link[1])
                new_temp_plugs.append(temp_plug)

            # To avoid the "_items" bug: setting the has_items attribute
            # of the trait's handler to False
            node = self.scene.pipeline.nodes['']
            source_trait = node.get_trait(plug_name)
            if source_trait.handler.has_items:
                source_trait.handler.has_items = False

            self.scene.pipeline.remove_trait(_temp_plug_name[1])
            self.scene.update_pipeline()

            # if not from_undo and not from_redo:
            if not from_export_plugs:
                # For history
                history_maker = ["remove_plug", _temp_plug_name,
                                 new_temp_plugs, optional]

                self.update_history(history_maker, from_undo, from_redo)

                self.main_window.statusBar().showMessage(
                    "Plug {0} has been removed.".format(_temp_plug_name[1]))

    def save_pipeline(self, filename=None):
        """Save the pipeline.

        :return: the pipeline file name
        """
        config = Config()
        if not filename:
            pipeline = self.scene.pipeline
            folder = os.path.abspath(os.path.join(config.get_mia_path(),
                                                  'processes',
                                                  'User_processes'))

            if not os.path.isdir(folder):
                os.mkdir(folder)

            if not os.path.isfile(os.path.abspath(
                    os.path.join(folder, '__init__.py'))):
                with open(os.path.abspath(os.path.join(
                        folder, '__init__.py')), 'w'):
                    pass

            filename = QtWidgets.QFileDialog.getSaveFileName(
                None, 'Save the pipeline', folder,
                'Compatible files (*.py);; All (*)')[0]

            if not filename:  # save widget was cancelled by the user
                return ''

            if os.path.splitext(filename)[1] == '':  # which means no extension
                filename += '.py'

            elif os.path.splitext(filename)[1] != '.py':
                msg = QtWidgets.QMessageBox()
                msg.setIcon(QtWidgets.QMessageBox.Warning)
                msg.setText('The pipeline will be saved with a' +
                            ' ".py" extension instead of {0}'.format(
                                os.path.splitext(filename)[1]))
                msg.setWindowTitle("Warning")
                msg.setStandardButtons(QtWidgets.QMessageBox.Ok)
                msg.buttonClicked.connect(msg.close)
                msg.exec()
                filename = os.path.splitext(filename)[0] + '.py'

        if filename:
            posdict = dict([(key, (value.x(), value.y())) for key, value in
                            six.iteritems(self.scene.pos)])
            dimdict = dict([(key, (value[0], value[1])) for key, value in
                            six.iteritems(self.scene.dim)])  # add by Irmage OM

            pipeline.node_dimension = dimdict  # add by Irmage OM

            old_pos = pipeline.node_position
            pipeline.node_position = posdict
            # pipeline_tools.save_pipeline(pipeline, filename)
            save_pipeline(pipeline, filename)
            self._pipeline_filename = unicode(filename)
            pipeline.node_position = old_pos

            self.pipeline_saved.emit(filename)
            return filename


def find_filename(paths_list, packages_list, file_name):
    """
    Find the corresponding file name in the paths list of process_config.yml.

    :param paths_list: list of all the paths contained in process_config.yml
    :param packages_list: packages path
    :param file_name: name of the sub-pipeline
    :return: name of the corresponding file if it is found, else None
    """

    filenames = [file_name + '.py', file_name + '.xml']
    for filename in filenames:
        for path in paths_list:
            new_path = path
            for package in packages_list:
                new_path = os.path.join(new_path, package)

            # Making sure that the filename is found (has somme issues
            # with case sensitivity)
            for f in os.listdir(new_path):
                new_file = os.path.join(new_path, f)
                if os.path.isfile(new_file) and f.lower() == filename.lower():
                    return new_file


def get_path(name, dictionary, prev_paths=None):
    """Return the package path to the selected sub-pipeline.

    :param name: name of the sub-pipeline
    :param dictionary: package tree (read from process_config.yml)
    :param prev_paths: paths of the last call of this function
    :return: the package path of the sub-pipeline if it is found, else None
    """

    if prev_paths is None:
        prev_paths = []

    # new_paths is a list containing the packages to the desired module
    new_paths = prev_paths.copy()
    for idx, (key, value) in enumerate(dictionary.items()):
        # If the value is a string, this means
        # that this is a "leaf" of the tree
        # so the key is a module name.
        if isinstance(value, str):
            if key == name:
                new_paths.append(key)
                return new_paths
            else:
                continue
        # Else, this means that the value is still a dictionary,
        # we are still in the tree
        else:
            new_paths.append(key)
            final_res = get_path(name, value, new_paths)
            # final_res is None if the module name has not
            # been found in the tree
            if final_res:
                return final_res
            else:
                new_paths = prev_paths.copy()


def save_pipeline(pipeline, filename):
    """Save the pipeline either in XML or .py source file."""
    formats = {'.py': save_py_pipeline,
               '.xml': save_xml_pipeline}
    saved = False

    for ext, writer in six.iteritems(formats):

        if filename.endswith(ext):
            writer(pipeline, filename)
            saved = True
            break

    if not saved:
        # fallback to .py
        save_py_pipeline(pipeline, filename)









# -*- coding: utf-8 -*- #

""" Handle the pipeline manager tab. Allow to create and run pipelines.

Contains:
    Module:
        - iteration_table
        - node_controller
        - pipeline_editor
        - pipeline_manager_tab
        - process_library
        - process_mia

"""

##########################################################################
# Populse_mia - Copyright (C) IRMaGe/CEA, 2018
# Distributed under the terms of the CeCILL license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL_V2.1-en.html
# for details.
##########################################################################



##########################################################################
# Populse_mia - Copyright (C) IRMaGe/CEA, 2018
# Distributed under the terms of the CeCILL license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL_V2.1-en.html
# for details.
##########################################################################

import sys
import sip
import six
import os
from matplotlib.backends.qt_compat import QtWidgets
from functools import partial
from traits.api import Undefined, TraitError

# PyQt5 imports
from PyQt5.QtGui import QIcon
from PyQt5.QtCore import pyqtSignal
from PyQt5.QtWidgets import (
    QPushButton, QVBoxLayout, QWidget, QHBoxLayout, QLabel, QLineEdit,
    QGroupBox, QMessageBox, QToolButton, QDialog, QDialogButtonBox,
    QApplication)

# soma-base imports
from soma.controller import trait_ids

# Populse_MIA imports
from populse_mia.user_interface.data_browser.advanced_search import (
    AdvancedSearch)
from populse_mia.user_interface.data_browser.rapid_search import RapidSearch
from populse_mia.user_interface.data_browser.data_browser import (
    TableDataBrowser, not_defined_value)
from populse_mia.user_interface.pop_ups import (
    PopUpSelectTagCountTable, PopUpVisualizedTags)
from populse_mia.data_manager.project import TAG_FILENAME, COLLECTION_CURRENT
from populse_mia.data_manager.filter import Filter
from populse_mia.user_interface.pipeline_manager.process_mia import ProcessMIA
from populse_mia.software_properties import Config

if sys.version_info[0] >= 3:
    unicode = str


    def values(d):
        return list(d.values())
else:
    def values(d):
        return d.values()


class NodeController(QWidget):
    """
    Allow to change the input and output values of a pipeline node

    Methods:
        - clearLayout: clear the layouts of the widget
        - display_filter: display a filter widget
        - display_parameters: display the parameters of the selected node
        - get_index_from_plug_name: return the index of the plug label
        - update_node_name: update the name of the selected node
        - update_parameters: update the parameters values
        - update_plug_value: update the value of a node plug
        - update_plug_value_from_filter: update the plug value from a filter
           result

    """

    value_changed = pyqtSignal(list)

    def __init__(self, project, scan_list, pipeline_manager_tab, main_window):
        """
        Initialization of the Node Controller

        :param project: current project in the software
        :param scan_list: list of the selected database files
        :param pipeline_manager_tab: parent widget
        :param main_window: main window of the software
        """

        super(NodeController, self).__init__(pipeline_manager_tab)

        self.project = project
        self.scan_list = scan_list
        self.main_window = main_window
        self.node_name = ""

        # Layouts
        self.v_box_final = QVBoxLayout()
        self.h_box_node_name = QHBoxLayout()

    def clearLayout(self, layout):
        """
        Clear the layouts of the widget

        :param layout: widget with a layout
        :return:
        """
        for i in reversed(range(len(layout.children()))):
            if type(layout.layout().itemAt(i)) == QtWidgets.QWidgetItem:
                layout.layout().itemAt(i).widget().setParent(None)
            if type(layout.layout().itemAt(i)) == QtWidgets.QHBoxLayout or \
                    type(layout.layout().itemAt(i)) == QtWidgets.QVBoxLayout:
                layout.layout().itemAt(i).deleteLater()
                for j in reversed(range(len(layout.layout().itemAt(i)))):
                    layout.layout().itemAt(i).itemAt(j).widget().setParent(
                        None)

        if layout.layout() is not None:
            sip.delete(layout.layout())

    def display_filter(self, node_name, plug_name, parameters, process):
        """
        Display a filter widget

        :param node_name: name of the node
        :param plug_name: name of the plug
        :param parameters: tuple containing the index of the plug, the current
           pipeline instance and the type of the plug value
        :param process: process of the node
        """
        self.pop_up = PlugFilter(self.project, self.scan_list, process,
                                 node_name, plug_name, self, self.main_window)
        self.pop_up.show()
        self.pop_up.plug_value_changed.connect(
            partial(self.update_plug_value_from_filter, plug_name, parameters))

    def display_parameters(self, node_name, process, pipeline):
        """
        Display the parameters of the selected node

        The node parameters are read and line labels/line edits/push buttons
        are created for each of them. This methods consists mainly in widget
        and layout organization.

        :param node_name: name of the node
        :param process: process of the node
        :param pipeline: current pipeline
        """

        self.node_name = node_name
        self.current_process = process

        self.line_edit_input = []
        self.line_edit_output = []
        self.labels_input = []
        self.labels_output = []

        # Refreshing the layouts
        if len(self.children()) > 0:
            self.clearLayout(self)

        self.v_box_final = QVBoxLayout()

        # Node name
        label_node_name = QLabel()
        label_node_name.setText('Node name:')

        self.line_edit_node_name = QLineEdit()

        # The pipeline global inputs and outputs node name cannot be modified
        if self.node_name not in ('inputs', 'outputs'):
            self.line_edit_node_name.setText(self.node_name)
            self.line_edit_node_name.returnPressed.connect(
                partial(self.update_node_name, pipeline))
        else:
            self.line_edit_node_name.setText('Pipeline inputs/outputs')
            self.line_edit_node_name.setReadOnly(True)

        self.h_box_node_name = QHBoxLayout()
        self.h_box_node_name.addWidget(label_node_name)
        self.h_box_node_name.addWidget(self.line_edit_node_name)

        # Inputs
        self.button_group_inputs = QGroupBox('Inputs')
        self.v_box_inputs = QVBoxLayout()
        idx = 0

        for name, trait in process.user_traits().items():
            if name == 'nodes_activation':
                continue
            if not trait.output:
                label_input = QLabel()
                label_input.setText(str(name))
                self.labels_input.insert(idx, label_input)
                try:
                    value = getattr(process, name)
                except TraitError:
                    value = Undefined
                trait_type = trait_ids(process.trait(name))

                self.line_edit_input.insert(idx, QLineEdit())
                self.line_edit_input[idx].setText(str(value))
                self.line_edit_input[idx].returnPressed.connect(
                    partial(self.update_plug_value, 'in', name, pipeline,
                            type(value)))

                h_box = QHBoxLayout()
                h_box.addWidget(label_input)
                h_box.addWidget(self.line_edit_input[idx])

                # Adding the possibility to filter pipeline global
                # inputs except if the input is "database_scans"
                # which means that the scans will be filtered with InputFilter
                if self.node_name == "inputs" and name != "database_scans":
                    parameters = (idx, pipeline, type(value))
                    push_button = QPushButton('Filter')
                    push_button.clicked.connect(
                        partial(self.display_filter, self.node_name, name,
                                parameters, process))
                    h_box.addWidget(push_button)

                self.v_box_inputs.addLayout(h_box)

                idx += 1

        self.button_group_inputs.setLayout(self.v_box_inputs)

        # Outputs
        self.button_group_outputs = QGroupBox('Outputs')
        self.v_box_outputs = QVBoxLayout()
        idx = 0
        for name, trait in process.traits(output=True).items():
            label_output = QLabel()
            label_output.setText(str(name))
            self.labels_output.insert(idx, label_output)

            value = getattr(process, name)
            trait_type = trait_ids(process.trait(name))

            self.line_edit_output.insert(idx, QLineEdit())
            self.line_edit_output[idx].setText(str(value))
            self.line_edit_output[idx].returnPressed.connect(
                partial(self.update_plug_value, 'out', name, pipeline,
                        type(value)))

            h_box = QHBoxLayout()
            h_box.addWidget(label_output)
            h_box.addWidget(self.line_edit_output[idx])

            self.v_box_outputs.addLayout(h_box)

            idx += 1

        self.button_group_outputs.setLayout(self.v_box_outputs)

        self.v_box_final.addLayout(self.h_box_node_name)
        self.v_box_final.addWidget(self.button_group_inputs)
        self.v_box_final.addWidget(self.button_group_outputs)

        self.setLayout(self.v_box_final)

    def get_index_from_plug_name(self, plug_name, in_or_out):
        """
        Return the index of the plug label.

        :param plug_name: name of the plug
        :param in_or_out: "in" if the plug is an input plug, "out" else
        :return: the corresponding index
        """
        if in_or_out == "in":
            for idx, label in enumerate(self.labels_input):
                if label.text() == plug_name:
                    return idx

        else:
            for idx, label in enumerate(self.labels_output):
                if label.text() == plug_name:
                    return idx

    def update_node_name(self, pipeline, new_node_name=None):
        """
        Change the name of the selected node and updates the pipeline.

        Because the nodes are stored in a dictionary, we have to create
        a new node that has the same traits as the selected one and create
        new links that are the same than the selected node.

        :param pipeline: current pipeline
        :param new_node_name: new node name (is None except when this method
           is called from an undo/redo)
        """

        # Copying the old node
        old_node = pipeline.nodes[self.node_name]
        old_node_name = self.node_name

        if not new_node_name:
            new_node_name = self.line_edit_node_name.text()

        if new_node_name in list(pipeline.nodes.keys()):
            print("Node name already in pipeline")

        else:
            # Removing links of the selected node and copy
            # the origin/destination
            links_to_copy = []
            for source_parameter, source_plug \
                    in six.iteritems(old_node.plugs):
                for (dest_node_name, dest_parameter, dest_node, dest_plug,
                     weak_link) in source_plug.links_to.copy():
                    pipeline.remove_link(
                        self.node_name + "." + source_parameter + "->" +
                        dest_node_name + "." + dest_parameter)
                    links_to_copy.append(("to", source_parameter,
                                          dest_node_name, dest_parameter))

                for (dest_node_name, dest_parameter, dest_node, dest_plug,
                     weak_link) in source_plug.links_from.copy():
                    pipeline.remove_link(
                        dest_node_name + "." + dest_parameter + "->" +
                        self.node_name + "." + source_parameter)
                    links_to_copy.append(("from", source_parameter,
                                          dest_node_name, dest_parameter))

            # Creating a new node with the new name and deleting the former
            pipeline.nodes[new_node_name] = pipeline.nodes[self.node_name]
            del pipeline.nodes[self.node_name]

            # Updating the node_name attribute
            self.node_name = new_node_name

            # Setting the same links as the original node
            for link in links_to_copy:

                if link[0] == "to":
                    pipeline.add_link(self.node_name + "." + link[1] + "->"
                                      + link[2] + "." + link[3])
                elif link[0] == "from":
                    pipeline.add_link(link[2] + "." + link[3] + "->"
                                      + self.node_name + "." + link[1])

            # Updating the pipeline
            pipeline.update_nodes_and_plugs_activation()

            # To undo/redo
            self.value_changed.emit(["node_name",
                                     pipeline.nodes[new_node_name],
                                     new_node_name, old_node_name])

            self.main_window.statusBar().showMessage(
                'Node name "{0}" has been changed to "{1}".'.format(
                    old_node_name, new_node_name))

    def update_parameters(self, process=None):
        """
        Update the parameters values

        :param process: process of the node
        """

        if process is None:
            try:
                process = self.current_process
            except AttributeError:
                # if no node has been clicked, no need to update the widget
                return

        idx = 0
        for name, trait in process.user_traits().items():
            if name == 'nodes_activation':
                continue
            if not trait.output:
                try:
                    value = getattr(process, name)
                except TraitError:
                    value = Undefined
                self.line_edit_input[idx].setText(str(value))
                idx += 1

        idx = 0

        for name, trait in process.traits(output=True).items():
            value = getattr(process, name)
            self.line_edit_output[idx].setText(str(value))
            idx += 1

    def update_plug_value(self, in_or_out, plug_name, pipeline, value_type,
                          new_value=None):
        """
        Update the value of a node plug

        :param in_or_out: "in" if the plug is an input plug, "out" else
        :param plug_name: name of the plug
        :param pipeline: current pipeline
        :param value_type: type of the plug value
        :param new_value: new value for the plug (is None except when this
           method is called from an undo/redo)
        """

        index = self.get_index_from_plug_name(plug_name, in_or_out)

        # Reading the value from the plug's line edit
        if not new_value:

            if in_or_out == 'in':
                new_value = self.line_edit_input[index].text()
            elif in_or_out == 'out':
                new_value = self.line_edit_output[index].text()
            else:
                new_value = None

            try:
                new_value = eval(new_value)
            except NameError:
                pass
            except Exception as err:
                print("{0}: {1}.".format(err.__class__, err))

            # except NameError:
            #    print("NameError for value {0}".format(new_value))
            # except SyntaxError:
            #    pass
            #    print("SynthaxError for value {0}".format(new_value))

            if value_type not in [float, int, str, list]:
                value_type = str

        if self.node_name in ['inputs', 'outputs']:
            node_name = ''
        else:
            node_name = self.node_name

        old_value = pipeline.nodes[node_name].get_plug_value(plug_name)

        try:
            pipeline.nodes[node_name].set_plug_value(plug_name, new_value)
        except TraitError as err:
            msg = QMessageBox()
            msg.setText("{0}: {1}.".format(err.__class__, err))
            msg.setIcon(QMessageBox.Warning)
            msg.exec_()

            if in_or_out == 'in':
                self.line_edit_input[index].setText(str(old_value))
            elif in_or_out == 'out':
                self.line_edit_output[index].setText(str(old_value))
            return

        # Update pipeline to "propagate" the node value
        pipeline.update_nodes_and_plugs_activation()

        if in_or_out == 'in':
            self.line_edit_input[index].setText(str(new_value))
        elif in_or_out == 'out':
            self.line_edit_output[index].setText(str(new_value))

        # To undo/redo
        self.value_changed.emit(["plug_value", self.node_name, old_value,
                                 plug_name, value_type, new_value])

        self.main_window.statusBar().showMessage(
            'Plug "{0}" of node "{1}" has been changed to "{2}".'.format(
                plug_name, node_name, new_value))

    def update_plug_value_from_filter(self, plug_name, parameters,
                                      filter_res_list):
        """
        Update the plug value from a filter result

        :param plug_name: name of the plug
        :param parameters: tuple containing the index of the plug,
          the current pipeline instance and the type of the plug value
        :param filter_res_list: list of the filtered files
        """

        index = parameters[0]
        pipeline = parameters[1]
        value_type = parameters[2]

        # If the list contains only one element, setting
        # this element as the plug value
        len_list = len(filter_res_list)
        if len_list > 1:
            res = filter_res_list
        elif len_list == 1:
            res = filter_res_list[0]
        else:
            res = []

        self.update_plug_value("in", plug_name, pipeline, value_type, res)


class PlugFilter(QWidget):
    """Filter widget used on a node plug

    The widget displays a browser with the selected files of the database,
    a rapid search and an advanced search to filter these files. Once the
    filtering is done, the result (as a list of files) is set to the plug.

    Methods:
        - ok_clicked: set the new value to the node plug and closes the widget
        - update_tag_to_filter: update the tag to Filter
        - update_tags: update the list of visualized tags
        - reset_search_bar: reset the search bar of the rapid search
        - search_str: update the files to display in the browser
        - set_plug_value: emit a signal to set the file names to the node plug
    """

    plug_value_changed = pyqtSignal(list)

    def __init__(self, project, scans_list, process, node_name, plug_name,
                 node_controller, main_window):
        """
        Initialization of the PlugFilter widget

        :param project: current project in the software
        :param scans_list: list of database files to filter
        :param process: process instance of the selected node
        :param node_name: name of the current node
        :param plug_name: name of the selected node plug
        :param node_controller: parent node controller
        :param main_window: parent main window
        """

        super(PlugFilter, self).__init__(None)

        from populse_mia.user_interface.data_browser.rapid_search import \
            RapidSearch
        from populse_mia.data_manager.project import \
            COLLECTION_CURRENT

        self.project = project
        self.node_controller = node_controller
        self.main_window = main_window
        self.process = process
        self.plug_name = plug_name

        # If the filter is saved in the node plug (not the case now)
        # if hasattr(self.process, 'filters'):
        #    if self.plug_name in self.process.filters.keys():
        #         print("Already a filter for {0} plug of {1} process".format(
        #                 self.plug_name, self.process.name))
        #         # TODO: fill the advanced search with the corresponding
        #          filter:orphan:

        # Verifying that the scan names begin not with a "/" or a "\"
        if scans_list:
            scans_list_copy = []
            for scan in scans_list:
                scan_no_pfolder = scan.replace(self.project.folder, "")
                if scan_no_pfolder[0] in ["\\", "/"]:
                    scan_no_pfolder = scan_no_pfolder[1:]
                scans_list_copy.append(scan_no_pfolder)

            self.scans_list = scans_list_copy

        # If there is no element in scans_list, this means that all the scans
        # of the database needs to be taken into account
        else:
            self.scans_list = self.project.session.get_documents_names(
                COLLECTION_CURRENT)

        self.setWindowTitle("Filter - " + node_name + " - " + plug_name)

        # Graphical components
        self.table_data = TableDataBrowser(self.project, self,
                                           self.node_controller.visibles_tags,
                                           False, True, link_viewer=False)

        # Reducing the list of scans to selection
        all_scans = self.table_data.scans_to_visualize
        self.table_data.scans_to_visualize = self.scans_list
        self.table_data.scans_to_search = self.scans_list
        self.table_data.update_visualized_rows(all_scans)

        search_bar_layout = QHBoxLayout()

        self.rapid_search = RapidSearch(self)
        self.rapid_search.textChanged.connect(partial(self.search_str))

        sources_images_dir = Config().getSourceImageDir()
        self.button_cross = QToolButton()
        self.button_cross.setStyleSheet('background-color:rgb(255, 255, 255);')
        self.button_cross.setIcon(QIcon(os.path.join(sources_images_dir,
                                                     'gray_cross.png')))
        self.button_cross.clicked.connect(self.reset_search_bar)

        search_bar_layout.addWidget(self.rapid_search)
        search_bar_layout.addWidget(self.button_cross)

        self.advanced_search = AdvancedSearch(
            self.project, self, self.scans_list,
            self.node_controller.visibles_tags, from_pipeline=True)
        self.advanced_search.show_search()

        push_button_tags = QPushButton("Visualized tags")
        push_button_tags.clicked.connect(self.update_tags)

        self.push_button_tag_filter = QPushButton(TAG_FILENAME)
        self.push_button_tag_filter.clicked.connect(self.update_tag_to_filter)

        push_button_ok = QPushButton("OK")
        push_button_ok.clicked.connect(self.ok_clicked)

        push_button_cancel = QPushButton("Cancel")
        push_button_cancel.clicked.connect(self.close)

        # Layout
        buttons_layout = QHBoxLayout()
        buttons_layout.addWidget(push_button_tags)
        buttons_layout.addWidget(self.push_button_tag_filter)
        buttons_layout.addStretch(1)
        buttons_layout.addWidget(push_button_ok)
        buttons_layout.addWidget(push_button_cancel)

        main_layout = QVBoxLayout()
        main_layout.addLayout(search_bar_layout)
        main_layout.addWidget(self.advanced_search)
        main_layout.addWidget(self.table_data)
        main_layout.addLayout(buttons_layout)

        self.setLayout(main_layout)

        screen_resolution = QApplication.instance().desktop().screenGeometry()
        width, height = screen_resolution.width(), screen_resolution.height()
        self.setMinimumWidth(0.6 * width)
        self.setMinimumHeight(0.8 * height)

    def ok_clicked(self):
        """
        Set the new value to the node plug and closes the widget

        """

        # To use if the filters are set on plugs, which is not the case
        # if isinstance(self.process, ProcessMIA):
        #     (fields, conditions, values, links, nots) =
        #     self.advanced_search.get_filters(False)
        #
        #
        #     plug_filter = Filter(None, nots, values, fields,
        #     links, conditions, "")
        #     self.process.filters[self.plug_name] = plug_filter

        self.set_plug_value()
        self.close()

    def reset_search_bar(self):
        """
        Reset the search bar of the rapid search

        """
        self.rapid_search.setText("")
        self.advanced_search.rows = []
        self.advanced_search.show_search()

        # All rows reput
        old_scan_list = self.table_data.scans_to_visualize
        self.table_data.scans_to_visualize = self.scans_list
        self.table_data.scans_to_search = self.scans_list
        self.table_data.update_visualized_rows(old_scan_list)

    def search_str(self, str_search):
        """
        Update the files to display in the browser
        :param str_search: string typed in the rapid search
        """

        from project.project import COLLECTION_CURRENT, TAG_FILENAME
        from data_browser.data_browser import not_defined_value

        old_scan_list = self.table_data.scans_to_visualize

        # Every scan taken if empty search
        if str_search == "":
            return_list = self.table_data.scans_to_search
        else:
            # Scans with at least a not defined value
            if str_search == not_defined_value:
                filter = self.prepare_not_defined_filter(
                    self.project.session.get_shown_tags())

            # Scans matching the search
            else:
                filter = self.rapid_search.prepare_filter(
                    str_search, self.project.session.get_shown_tags(),
                    self.table_data.scans_to_search)

            generator = self.project.session.filter_documents(
                COLLECTION_CURRENT, filter)

            # Creating the list of scans
            return_list = [getattr(scan, TAG_FILENAME) for scan in generator]

        self.table_data.scans_to_visualize = return_list
        self.advanced_search.scans_list = return_list

        # Rows updated
        self.table_data.update_visualized_rows(old_scan_list)

    def set_plug_value(self):
        """
        Emit a signal to set the file names to the node plug

        """

        result_names = []
        points = self.table_data.selectedIndexes()

        # If the use has selected some items
        if points:
            for point in points:
                row = point.row()
                tag_name = self.push_button_tag_filter.text()
                # We get the FileName of the scan from the first row
                scan_name = self.table_data.item(row, 0).text()
                value = self.project.session.get_value(COLLECTION_CURRENT,
                                                       scan_name, tag_name)
                if tag_name == TAG_FILENAME:
                    value = os.path.abspath(os.path.join(self.project.folder,
                                                         value))
                result_names.append(value)
        else:
            filter = self.table_data.get_current_filter()
            for i in range(len(filter)):
                scan_name = filter[i]
                tag_name = self.push_button_tag_filter.text()
                value = self.project.session.get_value(COLLECTION_CURRENT,
                                                       scan_name, tag_name)
                if tag_name == TAG_FILENAME:
                    value = os.path.abspath(os.path.join(self.project.folder,
                                                         value))
                result_names.append(value)

        self.plug_value_changed.emit(result_names)

    def update_tag_to_filter(self):
        """
        Update the tag to Filter

        """

        popUp = PopUpSelectTagCountTable(self.project,
                                         self.node_controller.visibles_tags,
                                         self.push_button_tag_filter.text())
        if popUp.exec_():
            self.push_button_tag_filter.setText(popUp.selected_tag)

    def update_tags(self):
        """
        Update the list of visualized tags

        """

        dialog = QDialog()
        visualized_tags = PopUpVisualizedTags(
            self.project,
            self.node_controller.visibles_tags)
        layout = QVBoxLayout()
        layout.addWidget(visualized_tags)
        buttons_layout = QHBoxLayout()
        buttons = QDialogButtonBox(QDialogButtonBox.Ok |
                                   QDialogButtonBox.Cancel)
        buttons.accepted.connect(dialog.accept)
        buttons.rejected.connect(dialog.reject)
        buttons_layout.addWidget(buttons)
        layout.addLayout(buttons_layout)
        dialog.setLayout(layout)
        dialog.show()
        dialog.setMinimumHeight(600)
        dialog.setMinimumWidth(600)
        if dialog.exec():
            new_visibilities = []
            for x in range(visualized_tags.list_widget_selected_tags.count()):
                visible_tag = visualized_tags.list_widget_selected_tags.item(
                    x).text()
                new_visibilities.append(visible_tag)
            new_visibilities.append(TAG_FILENAME)
            self.table_data.update_visualized_columns(
                self.node_controller.visibles_tags, new_visibilities)
            self.node_controller.visibles_tags = new_visibilities
            for row in self.advanced_search.rows:
                fields = row[2]
                fields.clear()
                for visible_tag in new_visibilities:
                    fields.addItem(visible_tag)
                fields.model().sort(0)
                fields.addItem("All visualized tags")


class FilterWidget(QWidget):
    """
    Filter widget used on a Input_Filter process

    The widget displays a browser with the selected files of the database,
    a rapid search and an advanced search to filter these files. Once the
    filtering is done, the filter is saved in the process.

    Attributes:
        - advanced_search: advanced search widget to build a complex filter
        - main_window: parent main window
        - process: process instance of the selected node
        - project: current project in the software
        - node: instance of the corresponding Input_Filter node
        - node_controller: parent node controller
        - rapid_search: rapid search widget to build a simple filter
        - scan_list: list of database files to filter
        - table_data: browser that displays the database files to filter
        - visible_tags: tags that are visible in the current project

    Methods:
        - update_tag_to_filter: update the tag to Filter
        - update_tags: update the list of visualized tags
        - reset_search_bar: reset the search bar of the rapid search
        - search_str: update the files to display in the browser
        - ok_clicked: sets the filter to the process and closes the widget
        - set_plug_value: sets the output of the filter to the output of the
           node
    """

    def __init__(self, project, node_name, node, main_window):
        """
        Initialization of the Filter Widget

        :param project: current project in the software
        :param node_name: name of the current node
        :param node: instance of the corresponding Input_Filter node
        :param main_window: parent main window
        """

        super(FilterWidget, self).__init__(None)

        self.setWindowTitle("Filter - " + node_name)
        self.project = project
        self.visible_tags = self.project.session.get_shown_tags()
        self.node = node
        self.process = node.process
        self.main_window = main_window

        scan_list = []
        # The scan list to filter corresponds to the input of the Input Filter
        if self.process.input and self.process is not Undefined:
            for scan in self.process.input:
                path, file_name = os.path.split(scan)
                path, second_folder = os.path.split(path)
                first_folder = os.path.basename(path)
                database_file = os.path.join(first_folder, second_folder,
                                             file_name)
                scan_list.append(database_file)

        self.scan_list = scan_list

        # Graphical components
        self.table_data = TableDataBrowser(self.project, self,
                                           self.visible_tags, False, False)

        # Reducing the list of scans to selection
        all_scans = self.table_data.scans_to_visualize

        self.table_data.scans_to_visualize = self.scan_list
        self.table_data.scans_to_search = self.scan_list
        self.table_data.update_visualized_rows(all_scans)

        # Filter information
        filter_to_apply = node.process.filter

        # Search
        self.rapid_search = RapidSearch(self)
        if filter_to_apply.search_bar:
            self.rapid_search.setText(filter_to_apply.search_bar)
        self.rapid_search.textChanged.connect(partial(self.search_str))

        self.advanced_search = AdvancedSearch(self.project, self,
                                              self.scan_list,
                                              self.visible_tags,
                                              from_pipeline=True)
        self.advanced_search.show_search()
        self.advanced_search.apply_filter(filter_to_apply)

        # Initialize Qt objects
        self.button_cross = QToolButton()
        self.push_button_tag_filter = QPushButton(TAG_FILENAME)

        self.layout_view()

    def layout_view(self):

        sources_images_dir = Config().getSourceImageDir()

        self.button_cross.setStyleSheet('background-color:rgb(255, 255, 255);')
        self.button_cross.setIcon(QIcon(os.path.join(sources_images_dir,
                                                     'gray_cross.png')))
        self.button_cross.clicked.connect(self.reset_search_bar)

        search_bar_layout = QHBoxLayout()
        search_bar_layout.addWidget(self.rapid_search)
        search_bar_layout.addWidget(self.button_cross)

        push_button_tags = QPushButton("Visualized tags")
        push_button_tags.clicked.connect(self.update_tags)

        self.push_button_tag_filter.clicked.connect(self.update_tag_to_filter)

        push_button_ok = QPushButton("OK")
        push_button_ok.clicked.connect(self.ok_clicked)

        push_button_cancel = QPushButton("Cancel")
        push_button_cancel.clicked.connect(self.close)

        # Layout
        buttons_layout = QHBoxLayout()
        buttons_layout.addWidget(push_button_tags)
        buttons_layout.addWidget(self.push_button_tag_filter)
        buttons_layout.addStretch(1)
        buttons_layout.addWidget(push_button_ok)
        buttons_layout.addWidget(push_button_cancel)

        main_layout = QVBoxLayout()
        main_layout.addLayout(search_bar_layout)
        main_layout.addWidget(self.advanced_search)
        main_layout.addWidget(self.table_data)
        main_layout.addLayout(buttons_layout)

        self.setLayout(main_layout)

        screen_resolution = QApplication.instance().desktop().screenGeometry()
        width, height = screen_resolution.width(), screen_resolution.height()
        self.setMinimumWidth(0.6 * width)
        self.setMinimumHeight(0.8 * height)

    def ok_clicked(self):
        """
        Set the filter to the process and closes the widget

        """
        if isinstance(self.process, ProcessMIA):
            (fields, conditions,
             values, links, nots) = self.advanced_search.get_filters(False)
            filt = Filter(None, nots, values,
                          fields, links, conditions, self.rapid_search.text())
            self.process.filter = filt

        self.set_output_value()
        self.close()

    def reset_search_bar(self):
        """
        Reset the search bar of the rapid search

        """
        self.rapid_search.setText("")
        self.advanced_search.rows = []
        self.advanced_search.show_search()

        # All rows reput
        old_scan_list = self.table_data.scans_to_visualize
        self.table_data.scans_to_visualize = self.scan_list
        self.table_data.scans_to_search = self.scan_list
        self.table_data.update_visualized_rows(old_scan_list)

    def search_str(self, str_search):
        """
        Update the files to display in the browser

        :param str_search: string typed in the rapid search
        """

        old_scan_list = self.table_data.scans_to_visualize

        # Every scan taken if empty search
        if str_search == "":
            return_list = self.table_data.scans_to_search
        else:
            # Scans with at least a not defined value
            if str_search == not_defined_value:
                filter = self.prepare_not_defined_filter(
                    self.project.session.get_shown_tags())
            # Scans matching the search
            else:
                filter = self.rapid_search.prepare_filter(
                    str_search, self.project.session.get_shown_tags(),
                    old_scan_list)

            generator = self.project.session.filter_documents(
                COLLECTION_CURRENT, filter)

            # Creating the list of scans
            return_list = [getattr(scan, TAG_FILENAME) for scan in generator]

        self.table_data.scans_to_visualize = return_list
        self.advanced_search.scans_list = return_list

        # Rows updated
        self.table_data.update_visualized_rows(old_scan_list)

    def set_output_value(self):
        """
        Set the output of the filter to the output of the node

        """

        result_names = []
        filter = self.table_data.get_current_filter()
        for i in range(len(filter)):
            scan_name = filter[i]
            tag_name = self.push_button_tag_filter.text()
            value = self.project.session.get_value(COLLECTION_CURRENT,
                                                   scan_name, tag_name)
            if tag_name == TAG_FILENAME:
                value = os.path.abspath(os.path.join(self.project.folder,
                                                     value))
            result_names.append(value)

        result_files = []
        for result_name in result_names:
            full_path = os.path.abspath(os.path.join(self.project.folder,
                                                     result_name))
            result_files.append(full_path)

        self.node.set_plug_value("output", result_files)

    def update_tag_to_filter(self):
        """
        Update the tag to Filter

        """

        pop_up = PopUpSelectTagCountTable(self.project, self.visible_tags,
                                          self.push_button_tag_filter.text())
        if pop_up.exec_():
            self.push_button_tag_filter.setText(pop_up.selected_tag)

    def update_tags(self):
        """
        Update the list of visualized tags

        """

        dialog = QDialog()
        visualized_tags = PopUpVisualizedTags(self.project, self.visible_tags)
        layout = QVBoxLayout()
        layout.addWidget(visualized_tags)
        buttons_layout = QHBoxLayout()
        buttons = QDialogButtonBox(QDialogButtonBox.Ok |
                                   QDialogButtonBox.Cancel)
        buttons.accepted.connect(dialog.accept)
        buttons.rejected.connect(dialog.reject)
        buttons_layout.addWidget(buttons)
        layout.addLayout(buttons_layout)
        dialog.setLayout(layout)
        dialog.show()
        dialog.setMinimumHeight(600)
        dialog.setMinimumWidth(600)
        if dialog.exec():
            new_visibilities = []
            for x in range(visualized_tags.list_widget_selected_tags.count()):
                visible_tag = visualized_tags.list_widget_selected_tags.item(
                    x).text()
                new_visibilities.append(visible_tag)
            new_visibilities.append(TAG_FILENAME)
            self.table_data.update_visualized_columns(self.visible_tags,
                                                      new_visibilities)
            self.node_controller.visibles_tags = new_visibilities
            for row in self.advanced_search.rows:
                fields = row[2]
                fields.clear()
                for visible_tag in new_visibilities:
                    fields.addItem(visible_tag)
                fields.model().sort(0)
                fields.addItem("All visualized tags")
